import warnings
from typing import Optional

import numpy as np
import pandas as pd

from leaspy.exceptions import LeaspyDataInputError

from .abstract_dataframe_data_reader import AbstractDataframeDataReader
from .individual_data import IndividualData

__all__ = ["EventDataframeDataReader"]


class EventDataframeDataReader(AbstractDataframeDataReader):
    """
    Methods to convert :class:`pandas.DataFrame` to `Leaspy`-compliant data containers for event data only.

    Parameters
    ----------
    event_time_name: str
        Name of the columns in dataframe that contains the time of event
    event_bool_name: str
        Name of the columns in dataframe that contains if the event is censored of not

    Raises
    ------
    :exc:`.LeaspyDataInputError`
    """

    def __init__(
        self,
        *,
        event_time_name: str = "EVENT_TIME",
        event_bool_name: str = "EVENT_BOOL",
        nb_events: Optional[int] = None,
    ):
        super().__init__()
        self.event_time_name = event_time_name
        self.event_bool_name = event_bool_name
        self.nb_events = nb_events

    @staticmethod
    def _check_headers(columns: list[str]) -> None:
        """
        Check mendatory dataframe headers

        Parameters
        ----------
        columns: List[str]
            Names of the columns headers of the dataframe that contains patients information
        """

        missing_mandatory_columns = ["ID"] if "ID" not in columns else []
        if len(missing_mandatory_columns) > 0:
            raise LeaspyDataInputError(
                f"Your dataframe must have {missing_mandatory_columns} columns"
            )

    def _set_index(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Set the index suited for the type of information contained in the dataframe. Here, events are only indices provided by patient ID

        Parameters
        ----------
        df: pd.DataFrame
            Dataframe with patient information

        Returns
        -------
        df: pd.DataFrame
            Dataframe with the right index
        """
        return df.set_index(["ID"])

    def _clean_dataframe(
        self, df: pd.DataFrame, *, drop_full_nan: bool, warn_empty_column: bool
    ) -> pd.DataFrame:
        """
        Clean the dataframe that contains patient events information

        Parameters
        ----------
        df: pd.DataFrame
            Dataframe with patient information

        drop_full_nan: bool
            If set to True, raw full of nan are dropped

        warn_empty_column: bool
            If set to True, a warning is raise for columns full of nan


        Returns
        -------
        df: pd.DataFrame
            Dataframe with clean information
        """

        # [SPECIFIC] check_available_data
        df_event = df.copy(deep=True)

        # Assert events columns are the only one available
        assert (df_event.columns == [self.event_time_name, self.event_bool_name]).all()

        # Round
        df_event[self.event_time_name] = round(
            df_event[self.event_time_name], self.time_rounding_digits
        )  # avoid missing duplicates due to rounding errors

        # Check event time data good format
        if not (df_event[self.event_time_name] > 0).all():
            raise LeaspyDataInputError("Events must be above 0")

        # Check event bool good format
        if not np.array_equal(
            df_event[self.event_bool_name], df_event[self.event_bool_name].astype(int)
        ):
            raise LeaspyDataInputError(
                "Events must be stored in type int, with 0 equal to censored event"
            )
        df_event[self.event_bool_name] = df_event[self.event_bool_name].astype(int)
        # Assert one unique event per patient and group to drop duplicates
        if (
            not (
                df_event.groupby("ID")
                .nunique()[[self.event_time_name, self.event_bool_name]]
                .eq(1)
            )
            .all()
            .all()
        ):
            raise LeaspyDataInputError(
                "There must be only an unique event_time and an unique event_bool per patient"
            )
        df_event = df_event.groupby("ID").first()

        # Event must be empty to raise an error
        if len(df_event) == 0:
            raise LeaspyDataInputError(
                "Dataframe should have at least 1 feature or an event"
            )

        nb_events = df_event[self.event_bool_name].max()
        if not self.nb_events:
            if nb_events == 0:
                raise LeaspyDataInputError(
                    f"There are no event, please check your data or put the number of events"
                )
            self.nb_events = nb_events

        elif self.nb_events != nb_events:
            if nb_events == 0:
                warnings.warn(
                    "There were no event in the dataset but you try to predict one"
                )
            else:
                raise LeaspyDataInputError(
                    "The number of events you provided is different from the number of events available in the data"
                )

        return df_event

    def _load_individuals_data(
        self, subj: IndividualData, df_subj: pd.DataFrame
    ) -> None:
        """
        Convert information stored in a dataframe to information stored into IndividualData

        Parameters
        ----------
        subj: IndividualData
            One patient with her/his information, potentially empty

        df_subj: pd.DataFrame
            One patient with her/his information
        """
        if self.nb_events < 1:
            raise LeaspyDataInputError(
                "The number of event should be equal or greater than 1"
            )

        time_at_event = [df_subj[self.event_time_name].unique()[0]] * self.nb_events

        bool_at_event = [False] * self.nb_events
        event_subj = df_subj[self.event_bool_name].unique()[0]
        if event_subj != 0:
            bool_at_event[event_subj - 1] = True

        subj.add_event(time_at_event, bool_at_event)



import copy
import json
import os
import warnings
from collections.abc import Iterable
from typing import Union

import pandas as pd
import torch

from leaspy.exceptions import (
    LeaspyIndividualParamsInputError,
    LeaspyInputError,
    LeaspyTypeError,
)
from leaspy.utils.typing import DictParamsTorch, IDType, ParamType

from ..data import Data, Dataset

__all__ = ["Result"]


class Result:
    """
    Result object class.
    Used as logs by personalize algorithms & simulation algorithm.

    Parameters
    ----------
    data : :class:`.Data`
        Object containing the information of the individuals, 
        in particular the time-points :math:`(t_{i,j})` and the observations :math:`(y_{i,j})`.
    individual_parameters : :obj:`dict` [:obj:`str`, :class:`torch.Tensor`]
        Contains log-acceleration 'xi', time-shifts 'tau' & 'sources'
    noise_std : :obj:`float` or :class:`torch.FloatTensor`, optional (default None)
        Desired noise standard deviation level

    Attributes
    ----------
    data : :class:`.Data`
        Object containing the information of the individuals, 
        in particular the time-points :math:`(t_{i,j})` and the observations :math:`(y_{i,j})`.
    individual_parameters : :obj:`dict` [:obj:`str`, :class:`torch.Tensor`]
        Contains log-acceleration 'xi', time-shifts 'tau' & 'sources' (dictionary of `torch.Tensor`).
    ID_to_idx : :obj:`dict`
        The keys are the individual ID & the items are their respective ordered position in the data file given
        by the user. This order remains the same during the computation.
        Example - in Result.individual_parameters['xi'], the first element corresponds to the
        first patient in ID_to_idx.
    noise_std : :obj:`float` or :class:`torch.FloatTensor`
        Desired noise standard deviation level.
    """

    # TODO : Check consistency and ordering of subjects ID between Data and individual parameters io.
    def __init__(
        self, data: Data, individual_parameters: DictParamsTorch, noise_std=None
    ):
        self.data = data
        self.individual_parameters = individual_parameters
        self.ID_to_idx: dict[IDType, int] = {
            key: i for i, key in enumerate(data.individuals)
        }
        self.noise_std = noise_std

    # TODO : this method is used only once in plotting => delete it ?
    def get_torch_individual_parameters(
        self, ID: Union[IDType, list[IDType]] = None
    ) -> DictParamsTorch:
        """
        Getter function for the individual parameters.

        Parameters
        ----------
        ID : :obj:`str` or :obj:`list`[:obj:`str`], optional (default None)
            Contains the identifiers of the wanted subject.

        Returns
        -------
        :obj:`dict` [:obj:`str`, :class:`torch.Tensor`]
            Contains the individual parameters.
        """
        if ID is not None:
            if not isinstance(ID, list):
                if isinstance(ID, str) or not isinstance(ID, Iterable):
                    # If ID is not a Iterable (case where ID is a int) => convert into list
                    # If ID is a str => convert into list
                    ID = [ID]
                else:
                    raise LeaspyIndividualParamsInputError(
                        "Input argument 'ID' must be a single identifier or a list or identifiers!"
                    )

            list_idt = [self.ID_to_idx[id_patient] for id_patient in ID]
            ind_parameters = {
                key: value[list_idt]
                for key, value in self.individual_parameters.items()
            }
        else:
            ind_parameters = self.individual_parameters.copy()
        return ind_parameters

    # TODO: unit test & functional test
    def get_dataframe_individual_parameters(
        self, cofactors: Union[str, list[str]] = None
    ) -> pd.DataFrame:
        """
        Return the dataframe of the individual parameters.

        Each row corresponds to a subject. The columns correspond
        (in this order) to the subjects' ID, the individual parameters (one column per individual parameter) & the
        cofactors (one column per cofactor).

        Parameters
        ----------
        cofactors : :obj:`str` or :obj:`list`[:obj:`str`], optional (default None)
            Contains the cofactor(s) to join to the logs dataframe.

        Returns
        -------
        :class:`pandas.DataFrame`
            Contains for each patient his ID & his individual parameters (optional and his cofactors states)

        Notes
        -----
        The cofactors must be present in the leaspy data object stored into the .data attribute of the result instance.
        See the example.

        Examples
        --------
        Load a longitudinal multivariate dataset & the subjects' cofactors. Compute the individual parameters for this
        dataset & get the corresponding dataframe with the genetic APOE cofactor

        >>> import pandas as pd
        >>> from leaspy.api import Leaspy
        >>> from leaspy.algo import AlgorithmSettings
        >>> from leaspy.io.data import Data
        >>> from leaspy.io.logs.visualization import Plotter
        >>> leaspy_logistic = Leaspy('logistic')
        >>> data = Data.from_csv_file('data/my_leaspy_data.csv')  # replace with your own path!
        >>> genes_cofactors = pd.read_csv('data/genes_cofactors.csv')  # replace with your own path!
        >>> print(genes_cofactors.head())
                   ID      APOE4
        0  sub-HS0102          1
        1  sub-HS0112          0
        2  sub-HS0113          0
        3  sub-HS0114          1
        4  sub-HS0115          0

        >>> data.load_cofactors(genes_cofactors, ['GENES'])
        >>> model_settings = AlgorithmSettings('mcmc_saem', seed=0)
        >>> personalize_settings = AlgorithmSettings('mode_real', seed=0)
        >>> leaspy_logistic.fit(data, model_settings)
        >>> individual_results = leaspy_logistic.personalize(data, model_settings)
        >>> individual_results_df = individual_results.get_dataframe_individual_parameters('GENES')
        >>> print(individual_results_df.head())
                           tau        xi  sources_0  sources_1  APOE4
        ID
        sub-HS0102   70.329201  0.120465   5.969921  -0.245034      1
        sub-HS0112   95.156624 -0.692099   1.520273   3.477707      0
        sub-HS0113   74.900673 -1.769864  -1.222979   1.665889      0
        sub-HS0114   81.792763 -1.003620   1.021321   2.371716      1
        sub-HS0115   89.724648 -0.820971  -0.480975   0.741601      0
        """
        # Initialize patient dict with ID
        patient_dict = {"ID": list(self.ID_to_idx.keys())}

        # For each individual variable
        for variable_ind in list(self.individual_parameters.keys()):
            # Case tau / xi --> unidimensional
            if self.individual_parameters[variable_ind].shape[1] == 1:
                patient_dict[variable_ind] = (
                    self.individual_parameters[variable_ind].numpy().reshape(-1)
                )
            # Case sources --> multidimensional
            elif self.individual_parameters[variable_ind].shape[1] > 1:
                for dim in range(self.individual_parameters[variable_ind].shape[1]):
                    patient_dict[f"{variable_ind}_{dim}"] = (
                        self.individual_parameters[variable_ind][:, dim]
                        .numpy()
                        .reshape(-1)
                    )

        df_individual_parameters = pd.DataFrame(patient_dict).set_index("ID")

        # If you want to load cofactors too
        if cofactors is not None:
            if isinstance(cofactors, str):
                cofactors = [cofactors]

            cofactor_dict = {"ID": list(self.data.individuals.keys())}

            for cofactor in cofactors:
                cofactor_dict[cofactor] = [
                    self.data.individuals[idx].cofactors[cofactor]
                    for idx in cofactor_dict["ID"]
                ]

            df_cofactors = pd.DataFrame(cofactor_dict).set_index("ID")
            df_individual_parameters = df_individual_parameters.join(df_cofactors)

        return df_individual_parameters

    def save_individual_parameters_csv(
        self, path: str, idx: list[IDType] = None, cofactors=None, **args
    ):
        """
        Save the individual parameters in a csv format.

        Parameters
        ----------
        path : :obj:`str`
            The logs' path.
        idx : :obj:`list` [:obj:`str`], optional (default None)
            Contain the IDs of the selected subjects. If ``None``, all the subjects are selected.
        cofactors : :obj:`str` or :obj:`list` [:obj:`str`], optional (default None)
            Contains the cofactor(s) to join to the logs dataframe.
        **args
            Parameters to pass to :meth:`pandas.DataFrame.to_csv`.

        Notes
        -----
        The cofactors must be present in the leaspy data object stored into the :attr:`.data` attribute of the result instance.
        See the example.

        Examples
        --------
        Save the individual parameters of the twenty first subjects.

        >>> from leaspy.algo import AlgorithmSettings
        >>> from leaspy.api import Leaspy
        >>> from leaspy.io.data import Data
        >>> leaspy_logistic = Leaspy('logistic')
        >>> data = Data.from_csv_file('data/my_leaspy_data.csv') # replace with your own path!
        >>> genes_cofactors = pd.read_csv('data/genes_cofactors.csv')  # replace with your own path!
        >>> data.load_cofactors(genes_cofactors, ['GENES'])
        >>> model_settings = AlgorithmSettings('mcmc_saem', seed=0)
        >>> personalize_settings = AlgorithmSettings('mode_real', seed=0)
        >>> leaspy_logistic.fit(data, model_settings)
        >>> individual_results = leaspy_logistic.personalize(data, model_settings)
        >>> output_path = 'outputs/logistic_seed0-mode_real_seed0-individual_parameter.csv'
        >>> idx = list(individual_results.individual_parameters.keys())[:20]
        >>> individual_results.save_individual_parameters_csv(output_path, idx, cofactors='GENES')
        """
        self._check_folder_existence(path)

        df_individual_parameters = self.get_dataframe_individual_parameters(
            cofactors=cofactors
        )
        if idx:
            if not isinstance(idx, list):
                raise LeaspyIndividualParamsInputError(
                    "Input 'idx' must be a list, even if it contains only one element! "
                    f"You gave idx={idx} which is of type {type(idx)}."
                )
            df_individual_parameters = df_individual_parameters.loc[idx]
        df_individual_parameters.to_csv(path, index=True, **args)

    def save_individual_parameters_json(
        self, path: str, idx: list[IDType] = None, human_readable=None, **args
    ):
        """
        Save the individual parameters in a json format.

        Parameters
        ----------
        path : :obj:`str`
            The logs' path.
        idx : :obj:`list` [:obj:`str`], optional (default None)
            Contain the IDs of the selected subjects. If ``None``, all the subjects are selected.
        human_readable : Any, optional (default None) -->  TODO change to bool

            .. deprecated:: 1.0

                * If None (default): save as json file
                * If not None: call :meth:`.save_individual_parameters_torch`.
        **args
            Arguments to pass to json.dump.
            Default to: dict(indent=2)

        Raises
        ------
        :class:`NotADirectoryError`
            if parent directory of path does not exist.

        Examples
        --------
        Save the individual parameters of the twenty first subjects.

        >>> from leaspy.algo import AlgorithmSettings
        >>> from leaspy.api import Leaspy
        >>> from leaspy.io.data import Data
        >>> leaspy_logistic = Leaspy('logistic')
        >>> data = Data.from_csv_file('data/my_leaspy_data.csv')
        >>> model_settings = AlgorithmSettings('mcmc_saem', seed=0)
        >>> personalize_settings = AlgorithmSettings('mode_real', seed=0)
        >>> leaspy_logistic.fit(data, model_settings)
        >>> individual_results = leaspy_logistic.personalize(data, model_settings)
        >>> output_path = 'outputs/logistic_seed0-mode_real_seed0-individual_parameter.json'
        >>> idx = list(individual_results.individual_parameters.keys())[:20]
        >>> individual_results.save_individual_parameters_json(output_path, idx)
        """
        self._check_folder_existence(path)
        dump = self._get_dump(idx)
        if human_readable is not None:
            warnings.warn(
                "This parameter is deprecated! To save as a torch file, use the method "
                "'save_individual_parameters_torch'.",
                DeprecationWarning,
                stacklevel=2,
            )
            self.save_individual_parameters_torch(path, idx)
        else:
            # Default json.dump kwargs:
            args = {"indent": 2, **args}

            with open(path, "w") as fp:
                json.dump(dump, fp, **args)

    def save_individual_parameters_torch(
        self, path: str, idx: list[IDType] = None, **args
    ):
        """
        Save the individual parameters in a torch format.

        Parameters
        ----------
        path : :obj:`str`
            The logs' path.
        idx : :obj:`list` [:obj:`str`], optional (default None)
            Contain the IDs of the selected subjects. If ``None``, all the subjects are selected.
        **args
            Arguments to pass to torch.save.

        Raises
        ------
        :exc:`NotADirectoryError`
            if parent directory of path does not exist.

        Examples
        --------
        Save the individual parameters of the twenty first subjects.

        >>> from leaspy.algo import AlgorithmSettings
        >>> from leaspy.api import Leaspy
        >>> from leaspy.io.data import Data
        >>> leaspy_logistic = Leaspy('logistic')
        >>> data = Data.from_csv_file('data/my_leaspy_data.csv')
        >>> model_settings = AlgorithmSettings('mcmc_saem', seed=0)
        >>> personalize_settings = AlgorithmSettings('mode_real', seed=0)
        >>> leaspy_logistic.fit(data, model_settings)
        >>> individual_results = leaspy_logistic.personalize(data, model_settings)
        >>> output_path = 'outputs/logistic_seed0-mode_real_seed0-individual_parameter.pt'
        >>> idx = list(individual_results.individual_parameters.keys())[:20]
        >>> individual_results.save_individual_parameters_torch(output_path, idx)
        """
        self._check_folder_existence(path)
        dump = self._get_dump(idx)
        torch.save(dump, path, **args)

    @staticmethod
    def _check_folder_existence(path: str):
        """
        Checks whether the folder in the given file path exists.

        Parameters
        ----------
        path : :obj:`str`
            The file path to check. May include a directory component.

        Raises
        ------
        :exc:`NotADirectoryError`
            If the directory part of the path is non-empty and does not exist.
        """
        # Test path's folder existence (if path contain a folder)
        dir_path = os.path.dirname(path)
        if not (dir_path == "" or os.path.isdir(dir_path)):
            raise NotADirectoryError(
                f"Cannot save individual parameter at path {path}. The folder does not exist!"
            )

    def _get_dump(self, idx: list[IDType] = None):
        """
        Convert the individual_parameters attribute into a dictionary of list. The univariate parameters values
        like xi and tau are squeeze from shape (n_subjects, 1) to (n_subjects,).
        One can select only the wanted subject by specifying their ID with 'idx' parameter.

        Parameters
        ----------
        idx : :obj:`list`, optional (default None)
            Contains the ID of the selected subjects.

        Returns
        -------
        :obj:`dict`
            A dictionary where keys are parameter names and values are lists of parameter values,
            either as flat lists (for univariate parameters) or lists of lists (for multivariate ones).
        """
        dump: dict = copy.deepcopy(self.individual_parameters)
        # Ex: individual_parameters = {'param1': torch.tensor([[1], [2], [3]]), ...}

        # Select only the wanted subjects
        if idx is not None:
            if not isinstance(idx, list):
                raise LeaspyIndividualParamsInputError(
                    "Input 'idx' must be a list, even if it contains only one element! "
                    f"You gave idx={idx} which is of type {type(idx)}."
                )
            selected_id = [self.ID_to_idx[val] for val in idx]
            dump = {key: val[selected_id] for key, val in dump.items()}

        for key in dump.keys():
            if not isinstance(dump[key], list):
                # For multivariate parameter - like sources
                # convert tensor([[1, 2], [2, 3]]) into [[1, 2], [2, 3]]
                if dump[key].shape[1] == 2:
                    dump[key] = dump[key].tolist()
                # for univariate parameters - like xi & tau
                # convert tensor([[1], [2], [3]]) into [1, 2, 3] => squeeze it
                elif dump[key].shape[1] == 1:
                    dump[key] = dump[key].squeeze().tolist()
        return dump

    @classmethod
    def load_individual_parameters_from_csv(cls, path: str, *, verbose=True, **kwargs):
        """
        Load individual parameters from a csv.

        Parameters
        ----------
        path : :obj:`str`
            The file's path. The csv file musts contain two columns named 'tau' and 'xi'. If the individual parameters
            come from a multivariate model, it must also contain the columns 'sources_i' for i in [0, ..., n_sources].
        verbose : :obj:`bool` (default True)
            Whether to have verbose output or not
        **kwargs
            Parameters to pass to :func:`pandas.read_csv`.

        Returns
        -------
        :obj:`dict` [:obj:`str`, :class:`torch.Tensor`]
            A dictionary of torch.tensor which contains the individual parameters.

        Examples
        --------
        Load an individual parameters dictionary from a saved file.

        >>> from leaspy.io.outputs import Result
        >>> path = 'outputs/logistic_seed0-mode_real_seed0-individual_parameter.csv'
        >>> individual_parameters = Result.load_individual_parameters_from_csv(path)
        """
        df = pd.read_csv(path, **kwargs)
        if verbose:
            print("Load from csv file ... conversion to torch")
        return cls.load_individual_parameters_from_dataframe(df)

    @staticmethod
    def load_individual_parameters_from_dataframe(df: pd.DataFrame):
        """
        Load individual parameters from a :class:`pandas.DataFrame`.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            Must contain two columns named 'tau' and 'xi'. If the individual parameters come from a multivariate model,
            it must also contain the columns 'sources_i' for i in [0, ..., n_sources].

        Returns
        -------
        :obj:`dict`[:obj:`str`, :class:`torch.Tensor`]
            A dictionary of torch.tensor which contains the individual parameters.
        """
        df.columns = [header.lower() for header in df.columns]
        sources_index = ["sources" in header for header in df.columns]
        ind_param = {
            "tau": torch.tensor(df["tau"].values, dtype=torch.float32).view(-1, 1),
            "xi": torch.tensor(df["xi"].values, dtype=torch.float32).view(-1, 1),
        }
        if any(sources_index):
            ind_param["sources"] = torch.tensor(
                df.iloc[:, sources_index].values, dtype=torch.float32
            )
        return ind_param

    @staticmethod
    def load_individual_parameters_from_json(path: str, *, verbose=True, **kwargs):
        """
        Load individual parameters from a json file.

        Deprecated : also load torch files.

        Parameters
        ----------
        path : :obj:`str`
            The file's path.
        verbose : :obj:`bool` (default True)
            Whether to have verbose output or not
        **kwargs
            Parameters to pass to :func:`json.load`.

        Returns
        -------
        :obj:`dict` [:obj:`str`, :class:`torch.Tensor`]
            A dictionary of `torch.Tensor` which contains the individual parameters.

        Examples
        --------
        Load an individual parameters dictionary from a saved file.

        >>> from leaspy.io.outputs import Result
        >>> path = 'outputs/logistic_seed0-mode_real_seed0-individual_parameter.json'
        >>> individual_parameters = Result.load_individual_parameters_from_json(path)
        """
        # Test if file is a json file
        try:
            with open(path, "r") as f:
                individual_parameters = json.load(f, **kwargs)
            if verbose:
                print("Load from json file ... conversion to torch")

            for key in individual_parameters.keys():
                # Convert every list in torch.tensor
                individual_parameters[key] = torch.tensor(
                    individual_parameters[key], dtype=torch.float32
                )
                # If tensor is 1-dimensional tensor([1, 2, 3]) => reshape it in tensor([[1], [2], [3]])
                if individual_parameters[key].dim() == 1:
                    individual_parameters[key] = individual_parameters[key].view(-1, 1)
        # Else if it is a torch file
        except UnicodeDecodeError:
            warnings.warn(
                "To load a torch file, use the static method result `load_individual_parameters_from_torch`",
                DeprecationWarning,
                stacklevel=2,
            )

            individual_parameters = torch.load(path)  # load function from torch
            if verbose:
                print("Load from torch file")

        return individual_parameters

    @staticmethod
    def load_individual_parameters_from_torch(path: str, *, verbose=True, **kwargs):
        """
        Load individual parameters from a torch file.

        Parameters
        ----------
        path : :obj:`str`
            The file's path.
        verbose : :obj:`bool` (default True)
            Whether to have verbose output or not
        **kwargs
            Parameters to pass to :func:`torch.load`.

        Returns
        -------
        :obj:`dict` [:obj:`str`, :class:`torch.Tensor`]
            A dictionary of `torch.Tensor` which contains the individual parameters.

        Examples
        --------
        Load an individual parameters dictionary from a saved file.

        >>> from leaspy.io.outputs import Result
        >>> path = 'outputs/logistic_seed0-mode_real_seed0-individual_parameter.pt'
        >>> individual_parameters = Result.load_individual_parameters_from_torch(path)
        """
        if verbose:
            print("Load from torch file")
        individual_parameters = torch.load(path, **kwargs)
        for key, val in individual_parameters.items():
            if not isinstance(val, torch.Tensor):
                individual_parameters[key] = torch.tensor(val, dtype=torch.float32)
            if individual_parameters[key].ndim != 2:
                individual_parameters[key] = individual_parameters[key].unsqueeze(-1)
        return individual_parameters

    @classmethod
    def load_individual_parameters(cls, path_or_df, **kwargs):
        """
        Load individual parameters from a :class:`pandas.DataFrame`, a csv, a json file or a torch file.

        Parameters
        ----------
        path_or_df : str or :class:`pandas.DataFrame`
            The file's path or a DataFrame containing the individual parameters.
        **kwargs
            Keyword-arguments to be passed to the corresponding load function.

        Returns
        -------
        :obj:`dict` [:obj:`str`, :class:`torch.Tensor`]
            A dictionary of torch.tensor which contains the individual parameters.

        Raises
        ------
        :exc:`FileNotFoundError`
            if path is invalid
        """
        if isinstance(path_or_df, pd.DataFrame):
            return cls.load_individual_parameters_from_dataframe(path_or_df)
        elif isinstance(path_or_df, str):
            file_extension = os.path.splitext(path_or_df)[-1]
            if file_extension == ".csv":
                return cls.load_individual_parameters_from_csv(path_or_df, **kwargs)
            elif file_extension == ".json":
                return cls.load_individual_parameters_from_json(path_or_df, **kwargs)
            else:
                if file_extension not in (".pt", ".p"):
                    warnings.warn(
                        f"File extension not recognized (got '{file_extension}')."
                        "Trying to load with torch by default.",
                        RuntimeWarning,
                        stacklevel=2,
                    )
                return cls.load_individual_parameters_from_torch(path_or_df, **kwargs)
        else:
            raise LeaspyIndividualParamsInputError(
                "The given input must be a pandas.DataFrame or a string "
                "giving the path of the file containing the individual parameters!"
            )

    @classmethod
    def load_result(cls, data, individual_parameters, *, cofactors=None, **kwargs):
        """
        Load a `Result` class object from two file - one for the individual data & one for the individual parameters.

        Parameters
        ----------
        data : :obj:`str` or :class:`pandas.DataFrame` or :class:`.Data`
            The file's path or a DataFrame containing the features' scores.
        individual_parameters : :obj:`str` or :class:`pandas.DataFrame`
            The file's path or a DataFrame containing the individual parameters.
        cofactors : :obj:`str` or :class:`pandas.DataFrame`, optional (default None)
            The file's path or a DataFrame containing the individual cofactors.
            The ID must be in index! Thus, the shape is (n_subjects, n_cofactors).
        **kwargs
            Parameters to pass to `Result.load_individual_parameters` method.

        Returns
        -------
        :class:`Result`
            A Result class object which contains the individual parameters and the individual data.

        Examples
        --------
        Launch an individual parameters estimation, save it and reload it.

        >>> from leaspy.algo import AlgorithmSettings
        >>> from leaspy.io.outputs import Result
        >>> from leaspy.api import Leaspy
        >>> from leaspy.io.data import Data
        >>> leaspy_logistic = Leaspy('logistic')
        >>> data = Data.from_csv_file('data/my_leaspy_data.csv')
        >>> model_settings = AlgorithmSettings('mcmc_saem', seed=0)
        >>> personalize_settings = AlgorithmSettings('mode_real', seed=0)
        >>> leaspy_logistic.fit(data, model_settings)
        >>> individual_results = leaspy_logistic.personalize(data, model_settings)
        >>> path_data = 'data/my_leaspy_data.csv'
        >>> path_individual_parameters = 'outputs/logistic_seed0-mode_real_seed0-individual_parameter.json'
        >>> individual_results.data.to_dataframe().to_csv(path_data)
        >>> individual_results.save_individual_parameters_json(path_individual_parameters)
        >>> individual_parameters = Result.load_result(path_data, path_individual_parameters)
        """
        if isinstance(data, Data):
            pass
        elif isinstance(data, str):
            data = Data.from_csv_file(data)
        elif isinstance(data, pd.DataFrame):
            data = Data.from_dataframe(data)
        else:
            raise LeaspyTypeError(
                "The given `data` input must be a Data instance, a pandas.DataFrame "
                "or a string giving the path of the file containing the features' scores! "
                f"You gave an object of type {type(data)}"
            )

        if cofactors is not None:
            if isinstance(cofactors, str):
                cofactors_df = pd.read_csv(cofactors, dtype={"ID": str}).set_index("ID")
            elif isinstance(cofactors, pd.DataFrame):
                cofactors_df = cofactors.copy()
            else:
                raise LeaspyTypeError(
                    "The given `cofactors` input must be a pandas.DataFrame "
                    "or a string giving the path of the file containing the cofactors! "
                    f"You gave an object of type {type(cofactors)}"
                )
            data.load_cofactors(cofactors_df)

        individual_parameters = cls.load_individual_parameters(
            individual_parameters, **kwargs
        )
        return cls(data, individual_parameters)

    def get_error_distribution_dataframe(self, model, cofactors=None):
        """
        Get signed residual distribution per patient, per sub-score & per visit. Each residual is equal to the
        modeled data minus the observed data.

        Parameters
        ----------
        model : :class:`~.models.abstract_model.AbstractModel`
        cofactors : str, list [str], optional (default None)
            Contains the cofactors' names to be included in the DataFrame. By default, no cofactors are returned.
            If cofactors == "all", all the available cofactors are returned.

        Returns
        -------
        residuals_dataframe : :class:`pandas.DataFrame` with index ['ID', 'TIME']

        Examples
        --------
        Get mean absolute error per feature:

        >>> from leaspy.algo import AlgorithmSettings
        >>> from leaspy.api import Leaspy
        >>> from leaspy.io.data import Data
        >>> data = Data.from_csv_file("/my/data/path")
        >>> leaspy_logistic = Leaspy('logistic')
        >>> settings = AlgorithmSettings("mcmc_saem", seed=0)
        >>> leaspy_logistic.calibrate(data, settings)
        >>> settings = AlgorithmSettings("mode_real", seed=0)
        >>> results = leaspy_logistic.personalize(data, settings)
        >>> residuals_dataframe = results.get_error_distribution_dataframe(model)
        >>> residuals_dataframe.abs().mean()
        """
        residuals_dataset = Dataset(self.data)
        residuals_dataset.values = (
            model.compute_individual_tensorized(
                residuals_dataset.timepoints, self.individual_parameters
            )
            - residuals_dataset.values
        )
        residuals_dataframe = residuals_dataset.to_pandas()

        if cofactors is not None:
            if isinstance(cofactors, str):
                if cofactors == "all":
                    cofactors_list = self.data.cofactors
                else:
                    cofactors_list = [cofactors]
            elif isinstance(cofactors, list):
                cofactors_list = cofactors
            else:
                raise LeaspyTypeError(
                    "The given `cofactors` input must be a string or a list of strings! "
                    f"You gave an object of type {type(cofactors)}"
                )
            cofactors_df = (
                self.data.to_dataframe(cofactors=cofactors)
                .groupby("ID")
                .first()[cofactors_list]
            )
            residuals_dataframe = residuals_dataframe.join(cofactors_df)

        return residuals_dataframe

    ###############################################################
    # DEPRECATION WARNINGS
    # These following methods will be removed in a future release
    ###############################################################

    @staticmethod
    def get_cofactor_states(cofactors: list) -> list:
        """
        .. deprecated:: 1.0

        Given a list of string return the list of unique elements.

        Parameters
        ----------
        cofactors : list[str]
            Distribution list of the cofactors.

        Returns
        -------
        list
            Unique occurrences of the input vector.
        """
        warnings.warn("This method will soon be removed!", DeprecationWarning)

        result = set(cofactors)
        return sorted(result)

    @staticmethod
    def _get_parameter_name_and_dim(param: str):
        """
        Splits a parameter string into its base name and optional dimension.

        Parameters
        ----------
        param : str
            The parameter name, possibly including a numeric suffix indicating a dimension.

        Returns
        -------
        tuple
            A tuple `(name, dim)`, where `name` is the base parameter name as a string,
            and `dim` is the parsed integer dimension, or `None` if no valid dimension is found.

        Examples
        --------
        >>> _get_parameter_name_and_dim(`abc_def_34`)
        ('abc_def', 34)
        """
        param_short, *param_dim = param.rsplit("_", maxsplit=1)  # from right

        if param_dim:
            # we found a last "***_NNN", return this split if and only if NNN can be interpreted as an integer
            try:
                return param_short, int(param_dim[0])
            except Exception:
                pass

        return param, None

    def get_parameter_distribution(self, parameter: ParamType, cofactor=None):
        """
        .. deprecated:: 1.0

        Return the wanted parameter distribution (one distribution per covariate state).

        Parameters
        ----------
        parameter : str
            The wanted parameter's name (ex: 'xi', 'tau', ...).
            It can also be `sources_i` to only get the i-th dimension of multivariate `sources` parameter.
        cofactor : str, optional (default None)
            The wanted cofactor's name.

        Returns
        -------
        list[float] or dict[str, Any]

        Raises
        ------
        :exc:`.LeaspyIndividualParamsInputError`
            if unsupported individual parameters
        :exc:`.LeaspyInputError`
            if unknown cofactor

        Notes
        -----
        If ``cofactor is None``:
            * If the parameter is univariate => return a list the parameter's distribution:
                list[float]
            * If the parameter is multivariate => return a dictionary:
                {'parameter1': distribution of parameter variable 1, 'parameter2': ...}

        If ``cofactor is not None``:
            * If the parameter is univariate => return a dictionary:
                {'cofactor1': parameter distribution such that patient.covariate = covariate1, 'cofactor2': ...}
            * If the parameter is multivariate => return a dictionary:
                {'cofactor1': {'parameter1': ..., 'parameter2': ...}, 'cofactor2': {...}, ...}
        """
        warnings.warn("This method will soon be removed!", DeprecationWarning)

        param_short, param_dim = self._get_parameter_name_and_dim(parameter)
        parameter_distribution = self.individual_parameters[
            param_short
        ]  # torch.tensor class object
        # parameter_distribution is of size (N_subjects, N_dimension_of_parameter)

        if param_dim is not None:
            parameter_distribution = parameter_distribution[:, [param_dim]]

        # Check the tensor's dimension is <= 2
        p_ndim = parameter_distribution.ndimension()
        if p_ndim > 2:
            raise LeaspyIndividualParamsInputError(
                f"The chosen parameter {parameter} is a tensor "
                f"of dimension {p_ndim}: it should be <= 2!"
            )
        ##############################################
        # If there is no cofactor to take into account
        ##############################################
        if cofactor is None:
            # If parameter is 1-dimensional
            if parameter_distribution.shape[1] == 1:
                # return a list of length = N_subjects
                parameter_distribution = parameter_distribution.view(-1).tolist()
            # Else transpose it and split it in a dictionary
            else:
                # return {'parameter1': distribution of parameter variable 1, 'parameter2': ... }
                parameter_distribution = {
                    parameter + str(i): val
                    for i, val in enumerate(
                        parameter_distribution.transpose(0, 1).tolist()
                    )
                }
            return parameter_distribution

        ############################################################
        # If the distribution as asked for different cofactor values
        ############################################################
        # Check if the cofactor exist
        all_cofactors = self.data[0].cofactors.keys()
        if cofactor not in all_cofactors:
            raise LeaspyInputError(
                f"The cofactor '{cofactor}' do not exist. "
                f"Here are the available cofactors: {list(all_cofactors)}"
            )
        # Get possible covariate stats
        # cofactors = [_.cofactors[cofactor] for _ in self.data if _.cofactors[cofactor] is not None]
        cofactors = self.get_cofactor_distribution(cofactor)
        cofactor_states = self.get_cofactor_states(cofactors)

        # Initialize the result
        distributions = {}

        # If parameter 1-dimensional
        if parameter_distribution.shape[1] == 1:
            parameter_distribution = parameter_distribution.view(
                -1
            ).tolist()  # ex: [1, 2, 3]
            # Create one entry per cofactor state
            for p in cofactor_states:
                if p not in distributions.keys():
                    distributions[p] = []
                # For each covariate state, get parameter distribution
                for i, v in enumerate(parameter_distribution):
                    if self.data[i].cofactors[cofactor] == p:
                        distributions[p].append(v)
                        # return {'cofactor1': ..., 'cofactor2': ...}
        else:
            # Create one dictionary per cofactor state
            for p in cofactor_states:
                if p not in distributions.keys():
                    # Create one dictionary per parameter dimension
                    distributions[p] = {
                        parameter + str(i): []
                        for i in range(parameter_distribution.shape[1])
                    }
                # Fill these entries by the corresponding values of the corresponding subject
                for i, v in enumerate(parameter_distribution.tolist()):
                    if self.data[i].cofactors[cofactor] == p:
                        for j, key in enumerate(distributions[p].keys()):
                            distributions[p][key].append(v[j])
                            # return {'cofactor1': {'parameter1': .., 'parameter2': ..}, 'cofactor2': { .. }, .. }
        return distributions

    def get_cofactor_distribution(self, cofactor: str):
        """
        .. deprecated:: 1.0

        Get the list of the cofactor's distribution.

        Parameters
        ----------
        cofactor : str
            Cofactor's name

        Returns
        -------
        list
            Cofactor's distribution.
        """
        warnings.warn("This method will soon be removed!", DeprecationWarning)

        return [d.cofactors[cofactor] for d in self.data]

    def get_patient_individual_parameters(self, idx: IDType):
        """
        .. deprecated:: 1.0

        Get the dictionary of the wanted patient's individual parameters

        Parameters
        ----------
        idx : str
            ID of the wanted patient

        Returns
        -------
        dict[param_name:str, `torch.Tensor`]
            Patient's individual parameters
        """
        warnings.warn("This method will soon be removed!", DeprecationWarning)

        # indices = list(self.data.individuals.keys())
        # idx_number = int(
        #     np.where(np.array(indices) == idx)[0])
        idx_number = [
            idx_nbr for idx_nbr, idxx in self.data.iter_to_idx.items() if idxx == idx
        ][0]

        patient_dict = dict.fromkeys(self.individual_parameters.keys())

        for variable_ind in list(self.individual_parameters.keys()):
            patient_dict[variable_ind] = self.individual_parameters[variable_ind][
                idx_number
            ]

        return patient_dict



"""This module defines the `VariablesDAG` class used to represent the relationships between the variables of a model."""

from collections import defaultdict
from collections.abc import Mapping
from dataclasses import dataclass, field
from queue import SimpleQueue
from typing import (
    Mapping as TMapping,
)
from typing import Type

import torch

from leaspy.exceptions import LeaspyInputError
from leaspy.utils.filtered_mapping_proxy import FilteredMappingProxy

from .specs import (
    IndividualLatentVariable,
    VariableInterface,
    VariableName,
    VariablesToFrozenSet,
)

__all__ = ["VariablesDAG"]


@dataclass(frozen=True)
class VariablesDAG(Mapping):
    """
    Directed acyclic graph of symbolic variables used in a leaspy model with efficient topologically sorted bidirectional access.

    Parameters
    ----------
    variables : :class:`~typing.Mapping` [ :class:`~leaspy.variables.specs.VariableName`, :class:`~leaspy.variables.specs.VariableInterface`]
        A dictionary mapping variable names to their specification objects (e.g., `IndepVariable`, `LinkedVariable`, etc.).
    direct_ancestors : :class:`~leaspy.variables.specs.VariablesToFrozenSet`
        A dictionary mapping each variable name to the set of variable names it directly depends on (i.e., its parents).
        Use :meth:`~leaspy.variables.dag.VariablesDAG.from_dict()` class method to reach the natural dependencies of linked variables.

    Notes
    -----
    Internally, this class precomputes:

    * ``direct_children``: inverse of `direct_ancestors`, mapping each node to its immediate dependents.

    * ``sorted_variables_names``: a topological ordering of variable names (roots to leaves).
        
    * ``sorted_children``: all transitive children of a node in topological order.
        
    * ``sorted_ancestors``: all transitive ancestors of a node in topological order.
        
    * ``sorted_variables_by_type``: variables grouped and ordered by their Python class/type.

    In general the VariablesDAG is not a tree because the graph may not be totally connected and may have multiple roots.
    It is not a multi-tree either since there may be multiple directed paths between two nodes - e.g. `logistic_model = f[g, b(g), ...]`.
    However, we do assume that there is no cycle in the graph (not checked currently), which is equivalent to be topologically sortable.

    In order to improve the efficiency of the algorithms, a node-wise sorted children and ancestors mappings are computed.
    These mappings are useful to:

    - perform computations and caching of intermediate variable dependencies
    - quickly reset all dependent nodes upon a modification

    Finally, we do not store children nor ancestors in a specific node class to avoid cross-references in such nodes.

    To Do
    -----
    - pre-compute roots (no ancestors) and leaves (no children) as well?
    - stratify variables dictionary per variable class?

    References
    ----------
    https://en.wikipedia.org/wiki/Directed_acyclic_graph#Computational_problems

    Examples
    --------
    >>> from leaspy.variables import VariablesDAG
    >>> from leaspy.variables.specs import IndepVariable, LinkedVariable
    >>> d_vars = {
        "x": IndepVariable(),
        "y": LinkedVariable(lambda *, x: -x),
    }
    >>> dag = VariablesDAG.from_dict(d_vars)
    """

    variables: TMapping[VariableName, VariableInterface]
    """Dictionary of variable names to their corresponding variable specification objects."""

    direct_ancestors: VariablesToFrozenSet = field(repr=False)
    """Mapping of variable names to their direct ancestors."""

    # pre-computed data that only depend on frozen attributes of dataclass
    direct_children: VariablesToFrozenSet = field(init=False, repr=False, compare=False)
    """Mapping of variable names to their direct children."""

    sorted_variables_names: tuple[VariableName, ...] = field(
        init=False, repr=False, compare=False
    )
    """Tuple of variable names sorted in topological order (from roots to leaves)."""
    # path_matrix: torch.Tensor = field(init=False, repr=False, compare=False)  # big and useless?
    sorted_children: TMapping[VariableName, tuple[VariableName, ...]] = field(
        init=False, repr=False, compare=False
    )
    """Mapping of all children of a given variable in a topological order."""

    sorted_ancestors: TMapping[VariableName, tuple[VariableName, ...]] = field(
        init=False, repr=False, compare=False
    )
    """Mapping of all ancestor of a given variable in a topological order."""

    sorted_variables_by_type: TMapping[
        # a better type hint would be nice here: Type[cls] -> Mapping[VariableName, cls]
        Type[VariableInterface],
        TMapping[VariableName, VariableInterface],
    ] = field(init=False, repr=False, compare=False)
    """The sorted variables, but stratified per variable type, to easily access them."""

    @classmethod
    def from_dict(cls, input_dictionary: TMapping[VariableName, VariableInterface]):
        """
        Instantiate a new DAG of variables from a dictionary of variables.

        This method is using :class:`~leaspy.variables.specs.LinkedVariable` dependencies as direct ancestors.

        Parameters
        ----------
        input_dictionary : :class:`~typing.Mapping` [:class:`~leaspy.variables.specs.VariableName` , :class:`~leaspy.variables.specs.VariableInterface`]
            The dictionary to use to create the DAG.

        Returns
        -------
        :class:`VariablesDAG`
            A new instance of `VariablesDAG` with the computed direct ancestor relationships.

        """
        direct_ancestors = {
            variable_name: variable.get_ancestors_names()
            for variable_name, variable in input_dictionary.items()
        }
        return cls(input_dictionary, direct_ancestors=direct_ancestors)

    def __post_init__(self):
        """
        Post-initialization routine to compute and cache derived attributes for the DAG.
        """
        nodes = self._check_consistency_of_nodes()
        children = self._compute_direct_children(nodes)
        (
            sorted_variables_names,
            sorted_children,
            sorted_ancestors,
        ) = self._compute_topological_orders(children)
        d_types = self._stratify_variables(sorted_variables_names)
        # Cache all those pre-computations values while keeping a "frozen" dataclass
        object.__setattr__(self, "direct_children", children)
        object.__setattr__(self, "sorted_variables_names", sorted_variables_names)
        # object.__setattr__(self, "path_matrix", path_matrix)
        object.__setattr__(self, "sorted_children", sorted_children)
        object.__setattr__(self, "sorted_ancestors", sorted_ancestors)
        object.__setattr__(self, "sorted_variables_by_type", d_types)

    def _check_consistency_of_nodes(self) -> frozenset[VariableName]:
        """
        Ensure internal consistency between variables and their declared ancestors.

        Returns
        -------
        :obj:`frozenset`[:class:`~leaspy.variables.specs.VariableName`]
            The set of all variable names (i.e. DAG nodes) in the graph.

        Raises
        ------
        :exc:`ValueError`
            If there is a mismatch between the declared variable names and the keys in `direct_ancestors`.

        """
        nodes = frozenset(self.variables.keys())
        if nodes != self.direct_ancestors.keys():
            raise ValueError("Inconsistent nodes in dictionary of ancestors edges")
        self._raise_if_bad_nodes_in_edges(
            self.direct_ancestors, nodes, what="ancestors"
        )
        return nodes

    def _compute_direct_children(
        self, nodes: frozenset[VariableName]
    ) -> VariablesToFrozenSet:
        """
        Compute children for efficient bidirectional access.
        
        Parameters
        ----------
        nodes : :obj:`frozenset`[:class:`~leaspy.variables.specs.VariableName`]
            The set of all variable names in the DAG.

        Returns
        -------
        :class:`~leaspy.variables.specs.VariablesToFrozenSet`
            A dictionary mapping each variable name to a frozenset of its direct children.

        Raises
        ------
        :exc:`ValueError`
            If a node appears isolated—i.e., not listed as an ancestor or child in either mapping.
        """
        children = defaultdict(set)
        for child_name, set_ancestors in self.direct_ancestors.items():
            for ancestor in set_ancestors:
                children[ancestor].add(child_name)
        children = {var_name: frozenset(children[var_name]) for var_name in nodes}
        self._raise_if_left_alone_nodes(children, self.direct_ancestors)
        return children

    def _compute_topological_orders(
        self,
        children: VariablesToFrozenSet,
    ) -> tuple[
        tuple[VariableName, ...],
        TMapping[VariableName, tuple[VariableName, ...]],
        TMapping[VariableName, tuple[VariableName, ...]],
    ]:
        """
        Compute the topological order of variables and their transitive relationships.

        Parameters
        ----------
        children : :class:`~leaspy.variables.specs.VariablesToFrozenSet`
            Mapping from each variable to its direct children.

        Returns
        -------
        sorted_variables_names : :obj:`tuple`[:class:`~leaspy.variables.specs.VariableName`, ...]
            A tuple of variable names sorted in topological order (from roots to leaves).
        sorted_children : :class:`~typing.Mapping`[:class:`~leaspy.variables.specs.VariableName`, :obj:`tuple`:class:`~leaspy.variables.specs.VariableName`, ...]]
            A dictionary mapping each variable to a tuple of its transitive children,
            ordered according to the topological sort.
        sorted_ancestors : :class:`~typing.Mapping`[:class:`~leaspy.variables.specs.VariableName`, :obj:`tuple`[:class:`~leaspy.variables.specs.VariableName`, ...]]
            A dictionary mapping each variable to a tuple of its transitive ancestors,
            ordered according to the topological sort.
        """
        (
            sorted_variables_names,
            path_matrix,
        ) = self.compute_topological_order_and_path_matrix(
            children, self.direct_ancestors
        )
        sorted_children, sorted_ancestors = self.compute_sorted_children_and_ancestors(
            sorted_variables_names, path_matrix
        )
        return sorted_variables_names, sorted_children, sorted_ancestors

    def _stratify_variables(
        self,
        sorted_variables_names: tuple[VariableName, ...],
    ) -> TMapping[Type[VariableInterface], TMapping[VariableName, VariableInterface]]:
        """
        Stratification of variables, per variable type.

        Parameters
        ----------
        sorted_variables_names : :obj:`tuple`[:class:`~leaspy.variables.specs.VariableName`, ...]
            A topologically sorted tuple of variable names, typically from roots to leaves.

        Returns
        -------
        :class:`~typing.Mapping`[:obj:`Type`[:class:`~leaspy.variables.specs.VariableInterface`], :class:`~typing.Mapping`[:class:`~leaspy.variables.specs.VariableName`, :class:`~leaspy.variables.specs.VariableInterface`]]
            A dictionary mapping each variable class/type to a filtered, read-only mapping of
            variable names to variable instances of that type.
        """
        d_types = defaultdict(list)
        for var_name in sorted_variables_names:
            d_types[type(self.variables[var_name])].append(var_name)
        return {
            var_type: FilteredMappingProxy(self.variables, subset=tuple(l_vars_type))
            for var_type, l_vars_type in d_types.items()
        }

    def __iter__(self):
        """Iterates on keys in topological order (.keys(), .values() and .items() methods are automatically provided by `Mapping`)."""
        return iter(self.sorted_variables_names)

    def __len__(self) -> int:
        """Get number of nodes."""
        return len(self.variables)

    def __getitem__(self, variable_name: VariableName) -> VariableInterface:
        """Get the variable specifications."""
        return self.variables[variable_name]

    @staticmethod
    def _raise_if_bad_nodes_in_edges(
        d_edges: VariablesToFrozenSet,
        s_nodes: frozenset[VariableName],
        *,
        what: str,
    ) -> None:
        """
        Validate that all referenced nodes in edge mappings are known and that there are no self-references.

        Parameters
        ----------
        d_edges : :class:`~typing.Mapping`[:class:`~leaspy.variables.specs.VariableName`, :obj:`frozenset`[:class:`~leaspy.variables.specs.VariableName`]]
            A dictionary mapping each node to a frozenset of connected nodes (e.g., ancestors or children).
        s_nodes : :obj:`frozenset`[:class:`~leaspy.variables.specs.VariableName`]
            The set of valid nodes expected in the graph.
        what : :obj:`str`
            A descriptive label for the type of edges being checked (e.g., "ancestors", "children"),
            used in error messages.

        Raises
        ------
        :exc:`LeaspyInputError`
            If there are unknown nodes referenced in the edges, or if any self-loops are detected.
        """
        pooled_nodes_from_edges = set().union(*d_edges.values())
        unknown_nodes = pooled_nodes_from_edges.difference(s_nodes)
        if len(unknown_nodes):
            raise LeaspyInputError(
                f"Those {what} variables are unknown: {unknown_nodes}"
            )
        self_loops = {n for n, s_connected in d_edges.items() if n in s_connected}
        if len(self_loops):
            raise LeaspyInputError(f"Those variables have self {what}: {self_loops}")

    @staticmethod
    def _raise_if_left_alone_nodes(
        d_children: VariablesToFrozenSet,
        d_ancestors: VariablesToFrozenSet,
    ) -> None:
        """
        Raise an error if any node has neither children nor ancestors, indicating an isolated node.
        
        Parameters
        ----------
        d_children : :class:`~typing.Mapping`[:class:`~leaspy.variables.specs.VariableName`, :obj:`frozenset`[:class:`~leaspy.variables.specs.VariableName`]]
            Mapping from each variable to its direct children.

        d_ancestors ::class:`~typing.Mapping`[:class:`~leaspy.variables.specs.VariableName`, :obj:`frozenset`[:class:`~leaspy.variables.specs.VariableName`]]
            Mapping from each variable to its direct ancestors.

        Raises
        ------
        :exc:`LeaspyInputError`
            If any variables are detected that have no children and no ancestors, i.e., are left completely alone.
        """
        s_left_alone = {
            var_name
            for var_name, s_ancestors in d_children.items()
            if len(s_ancestors) == 0 and len(d_ancestors[var_name]) == 0
        }
        if len(s_left_alone):
            raise LeaspyInputError(
                f"There are some variables left alone: {s_left_alone}"
            )

    @staticmethod
    def compute_topological_order_and_path_matrix(
        direct_children: VariablesToFrozenSet,
        direct_ancestors: VariablesToFrozenSet,
    ) -> tuple[tuple[VariableName, ...], torch.Tensor]:
        """
        Produce a topological sorting of the DAG.

        This relies on a modified Kahn's algorithm to produce a topological sorting of DAG,
        and the corresponding path matrix as a by-product.

        Parameters
        ----------
        direct_children : :class:`~leaspy.variables.specs.VariablesToFrozenSet`
            Mapping from each variable to its direct children.
        direct_ancestors : :class:`~leaspy.variables.specs.VariablesToFrozenSet`
            Mapping from each variable to its direct ancestors.

        Returns
        -------
        sorted_nodes : :obj:`tuple` [:class:`~leaspy.variables.specs.VariableName`, ...]
            Nodes in a topological order.
        path_matrix : :class:`torch.Tensor` [:obj:`bool`]
            Boolean triangle superior (strict) matrix indicating whether
            there is a (directed) path between nodes.

        Notes
        -----
        The algorithm has linear complexity with the O(number of edges + number of nodes).
        Input nodes are sorted by name in order to have fully reproducible output
        of the initial order of nodes and edges.
        (Thus renaming nodes may change the output, due to non-uniqueness of topological order)
        """
        nodes = sorted(direct_ancestors.keys())
        if set(nodes) != direct_children.keys():
            raise ValueError(
                "The nodes in provided 'direct_ancestors' do not match "
                "the nodes in provided 'direct_children'."
            )
        n_nodes = len(nodes)
        ix_nodes = {n: i for i, n in enumerate(nodes)}
        # copy of direct_ancestors & direct_children, with fixed order of nodes
        direct_ancestors_ = {n: direct_ancestors[n] for n in nodes}
        direct_children_ = {n: sorted(direct_children[n]) for n in nodes}
        # from roots (no ancestors) to leaves (no children)
        sorted_nodes: tuple[VarName, ...] = ()
        # indices of matrix correspond to `ix_nodes` until topological order is found
        path_matrix = torch.zeros((n_nodes, n_nodes), dtype=torch.bool)
        q_roots = SimpleQueue()
        for n, s_ancestors in direct_ancestors_.items():
            if len(s_ancestors) == 0:
                q_roots.put(n)
        while not q_roots.empty():
            n = q_roots.get()
            sorted_nodes += (n,)
            i = ix_nodes[n]
            for m in direct_children_[n]:
                j = ix_nodes[m]
                path_matrix[:, j] |= path_matrix[:, i]
                path_matrix[i, j] = True
                # drop edge (of the local copy of edges); no need to drop in `direct_children_`
                direct_ancestors_[m] = direct_ancestors_[m].difference({n})
                if len(direct_ancestors_[m]) == 0:
                    q_roots.put(m)
        if set(sorted_nodes) != set(nodes):
            raise ValueError("Input graph is not a DAG")
        # reorder elements of path matrix before returning it
        ix_sorted_nodes = [ix_nodes[n] for n in sorted_nodes]
        path_matrix = path_matrix[ix_sorted_nodes, :][:, ix_sorted_nodes]
        if not torch.equal(path_matrix, path_matrix.triu(1)):
            raise ValueError(
                f"Input graph is not a DAG: sorted path matrix = {path_matrix}"
            )
        return sorted_nodes, path_matrix

    @staticmethod
    def compute_sorted_children_and_ancestors(
        sorted_nodes: tuple[VariableName, ...],
        path_matrix: torch.Tensor,
    ) -> tuple[
        dict[VariableName, tuple[VariableName, ...]],
        dict[VariableName, tuple[VariableName, ...]],
    ]:
        """
        Produce node-wise topologically sorted children and ancestors from provided nodes.

        Parameters
        ----------
        sorted_nodes : :obj:`tuple` of :class:`~leaspy.variables.specs.VariableName`
            The sorted nodes.
        path_matrix : :class:`torch.Tensor`
            A binary 2D tensor where `path_matrix[i, j] == 1` indicates a path from node `i` to node `j`.
            Assumes the row/column indices correspond to the order of `sorted_nodes`.

        Returns
        -------
        sorted_children : :obj:`dict` [:class:`~leaspy.variables.specs.VariableName`, :obj:`tuple` [:class:`~leaspy.variables.specs.VariableName`, ...]]
            The sorted children.
        sorted_ancestors : :obj:`dict` [:class:`~leaspy.variables.specs.VariableName`, :obj:`tuple` [:class:`~leaspy.variables.specs.VariableName`, ...]]
            The sorted ancestors.
        """
        sorted_children = {
            node: tuple(
                sorted_nodes[j]
                for j in path_matrix[idx_node, :].nonzero(as_tuple=False).squeeze(dim=1)
            )
            for idx_node, node in enumerate(sorted_nodes)
        }
        sorted_ancestors = {
            node: tuple(
                sorted_nodes[i]
                for i in path_matrix[:, idx_node].nonzero(as_tuple=False).squeeze(dim=1)
            )
            for idx_node, node in enumerate(sorted_nodes)
        }
        return sorted_children, sorted_ancestors

    @property
    def individual_variable_names(self) -> tuple[VariableName, ...]:
        """
        Returns a tuple of variable names corresponding to the individual variables.

        Returns
        -------
        :obj:`tuple` of :class:`~leaspy.variables.specs.VariableName` :
            The individual variable names.
        """
        try:
            return tuple(self.sorted_variables_by_type[IndividualLatentVariable].keys())
        except KeyError:
            return ()



import os
import warnings

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch

from leaspy.exceptions import (
    LeaspyIndividualParamsInputError,
    LeaspyInputError,
    LeaspyTypeError,
)

from ...outputs import IndividualParameters

__all__ = ["Plotting"]


# TODO: outdated -
class Plotting:
    """
    .. deprecated:: 1.2

    Class defining some plotting tools.

    Parameters
    ----------
    model : :class:`~leaspy.models.BaseModel`
        The used model.
    output_path : :obj:`str`, (optional)
        Folder where plots will be saved.
        If None, default to current working directory.
    palette : :obj:`str` (palette name) or :class:`matplotlib.colors.Colormap` (`ListedColormap` or `LinearSegmentedColormap`)
        The palette to use.
    max_colors : :obj:`int` > 0, optional (default, corresponding to model nb of features)
        Only used if palette is a string
    """

    def __init__(self, model, output_path=".", palette="tab10", max_colors=10):
        warnings.warn(
            "Plotting will soon be removed from Leaspy, please use Plotter instead.",
            FutureWarning,
        )

        self.model = model

        # ---- Graphical options
        self.color_palette = None
        self.standard_size = (8, 4)
        self.linestyle = {
            "average_model": "-",
            "individual_model": "-",
            "individual_data": "-",
        }
        self.linewidth = {
            "average_model": 5,
            "individual_model": 2,
            "individual_data": 2,
        }
        self.alpha = {"average_model": 0.5, "individual_model": 1, "individual_data": 1}
        self.output_path = output_path

        self.set_palette(palette, max_colors)

    def set_palette(self, palette, max_colors=None):
        """
        Set palette of plots

        Parameters
        ----------
        palette : :obj:`str` (palette name) or :class:`matplotlib.colors.Colormap` (`ListedColormap` or `LinearSegmentedColormap`)
            The palette to use.

        max_colors : :obj:`int` > 0, optional (default, corresponding to model nb of features)
            Only used if palette is a string
        """

        if isinstance(palette, mpl.colors.Colormap):
            self.color_palette = palette
        else:
            if max_colors is None:
                if self.model.dimension is not None:
                    raise LeaspyInputError(
                        "Initialize model first please, with a not None dimension"
                    )
                max_colors = self.model.dimension
            self.color_palette = mpl.colormaps[palette].resampled(max_colors)

    def colors(self, at=None):
        """
        Wrapper over color_palette iterator to get colors

        Parameters
        ----------
        at : any legit color_palette arg (int, float or iterable of any of these) or None (default)
            if None returns all colors of palette upto model dimension

        Returns
        -------
        colors : single color tuple (RGBA) or np.array of RGBA colors (number of colors x 4)
        """
        if at is None:
            at = [i % self.color_palette.N for i in range(self.model.dimension)]

        return self.color_palette(at)

    def _raise_if_model_not_init(self):
        # /!\ Break if model is not initialized
        if not self.model.is_initialized:
            raise LeaspyInputError("Please initialize the model before plotting")

    def _handle_kwargs_begin(self, kwargs, all_features_list=None):
        """Extract kwargs corresponding to plot information and remove associated keys (in-place)."""

        # get features from initialized model if not set
        if all_features_list is None:
            self._raise_if_model_not_init()
            all_features_list = self.model.features

        # ---- Get requested features (may be a subset)
        features = kwargs.pop("features", all_features_list)
        features_ix = list(map(all_features_list.index, features))

        # ---- Colors
        colors = kwargs.pop("color", self.colors(features_ix))
        if len(colors) < len(features):
            raise LeaspyInputError(
                f"Please choose a palette with at least {len(features)} colors."
            )
        # TODO: reindex default colors if subset of features?

        # ---- Labels
        labels = kwargs.pop("labels", features)
        if len(labels) != len(features):
            raise LeaspyInputError(
                f"Dimensions mismatch between features ({len(features)}) and labels ({len(labels)}."
            )

        # ---- Ax
        ax = kwargs.pop("ax", None)
        if ax is None:
            fig, ax = plt.subplots(
                1, 1, figsize=kwargs.pop("figsize", self.standard_size)
            )

        # ---- Handle ylim
        if "logistic" in self.model.name:
            ax.set_ylim(0, 1)

        return ax, features, features_ix, labels, colors

    def _handle_kwargs_end(self, ax, kwargs, colors, labels):
        # ---- Legend
        dimension = len(labels)
        # if dimension is None:
        #    dimension = self.model.dimension

        custom_lines = [
            mpl.lines.Line2D([0], [0], color=colors[i], lw=4) for i in range(dimension)
        ]
        ax.legend(custom_lines, labels, title="Features")
        # ax.legend(title='Features')
        ax.set_ylabel("Normalized score")

        # ---- Save
        if "save_as" in kwargs.keys():
            plt.savefig(os.path.join(self.output_path, kwargs["save_as"]))

    def average_trajectory(self, **kwargs):
        """
        Plot the population average trajectories. They are parametrized by the population parameters derived
        during the calibration.

        Parameters
        ----------
        **kwargs
            * alpha: :obj:`float`, default 0.6
                Matplotlib's transparency option. Must be in [0, 1].
            * linestyle: {'-', '--', '-.', ':', '', (offset, on-off-seq), ...}
                Matplotlib's linestyle option.
            * linewidth: :obj:`float`
                Matplotlib's linewidth option.
            * features: list[:obj:`str`]
                Name of features (if set it must be a subset of model features)
                Default: all model features.
            * colors: list[:obj:`str`]
                Contains matplotlib compatible colors.
                At least as many as number of features.
            * labels: list[:obj:`str`]
                Used to rename features in the plot.
                Exactly as many as number of features.
                Default: raw variable name of each feature
            * ax: matplotlib.axes.Axes
                Axes object to modify, instead of creating a new one.
            * figsize: tuple of int
                The figure's size.
            * save_as: :obj:`str`, default None
                Path to save the figure.
            * title: :obj:`str`
            * n_tpts: :obj:`int`
                Number of timepoints in plot (default: 100)
            * n_std_left, n_std_right: :obj:`float` (default: 3 and 6 resp.)
                Time window around `tau_mean`, expressed as times of max(`tau_std`, 4)

        Returns
        -------
        :class:`matplotlib.axes.Axes`
        """
        # ---- Input manager
        plot_kws = self._plot_kwargs("average", kwargs)

        ax, _, features_ix, labels, colors = self._handle_kwargs_begin(kwargs)

        # ---- Get timepoints
        mean_time = self.model.parameters["tau_mean"].item()
        std_time = max(self.model.parameters["tau_std"].item(), 4)
        timepoints = mean_time + std_time * np.linspace(
            -kwargs.get("n_std_left", 3),
            kwargs.get("n_std_right", 6),
            kwargs.get("n_tpts", 100),
        )
        timepoints = torch.tensor(timepoints, dtype=torch.float32).unsqueeze(0)

        # ---- Compute average trajectory
        mean_trajectory = (
            self.model.compute_mean_traj(timepoints).cpu().detach().numpy()
        )

        # ---- plot it for each dimension
        for ft_ix, ft_lbl, ft_color in zip(features_ix, labels, colors):
            ax.plot(
                timepoints[0, :].cpu().detach().numpy(),
                mean_trajectory[0, :, ft_ix],
                c=ft_color,
                # label=ft_lbl, # not needed
                **plot_kws["model"],
            )

        # ---- Title & labels
        ax.set_title("Average trajectories")
        ax.set_xlabel("Age")

        self._handle_kwargs_end(ax, kwargs, colors, labels)

        return ax

    def _plot_kwargs(self, case, kwargs):
        if case == "average":
            return {
                "model": dict(
                    alpha=kwargs.get("alpha", self.alpha["average_model"]),
                    linestyle=kwargs.get("linestyle", self.linestyle["average_model"]),
                    linewidth=kwargs.get("linewidth", self.linewidth["average_model"]),
                )
            }
        elif case == "obs":
            return {
                "obs": dict(
                    alpha=kwargs.get("alpha", self.alpha["individual_data"]),
                    linestyle=kwargs.get(
                        "linestyle", self.linestyle["individual_data"]
                    ),
                    linewidth=kwargs.get(
                        "linewidth", self.linewidth["individual_data"]
                    ),
                    marker=kwargs.get("marker", "o"),
                    markersize=kwargs.get("markersize", "3"),
                )
            }
        elif case == "recons":
            # both observations & model will be displayed
            p_obs = dict(
                marker=kwargs.get("marker", "o"),  # None not to display obs
                markersize=kwargs.get("markersize", "4"),
                alpha=kwargs.get("obs_alpha", self.alpha["individual_data"]),
                linestyle=kwargs.get("obs_ls", ""),
                linewidth=kwargs.get("obs_lw", self.linewidth["individual_data"]),
            )
            p_model = dict(
                alpha=kwargs.get("alpha", self.alpha["individual_model"]),
                linestyle=kwargs.get("linestyle", self.linestyle["individual_model"]),
                linewidth=kwargs.get("linewidth", self.linewidth["individual_model"]),
            )
            return {"obs": p_obs, "model": p_model}
        else:
            raise LeaspyInputError("case must be in {'average', 'obs', 'recons'}")

    @staticmethod
    def _get_ip_df_torch(individual_parameters):
        # convert individual parameters in different cases

        if isinstance(individual_parameters, IndividualParameters):
            ip_df = individual_parameters.to_dataframe()
            ip_torch = individual_parameters.to_pytorch()
        elif isinstance(individual_parameters, pd.DataFrame):
            ip_df = individual_parameters
            ip_torch = IndividualParameters.from_dataframe(
                individual_parameters
            ).to_pytorch()
        elif isinstance(individual_parameters, tuple):
            ip_df = IndividualParameters.from_pytorch(
                *individual_parameters
            ).to_dataframe()
            ip_torch = individual_parameters
        else:
            raise LeaspyTypeError(
                "`individual_parameters` should be an IndividualParameters object, a pandas.DataFrame or a dict."
            )

        if ip_df.index.names != ["ID"]:
            raise LeaspyIndividualParamsInputError(
                "Individual parameters index is not ['ID'] "
                f"as expected but {list(ip_df.index.names)}"
            )

        return ip_df, ip_torch

    def _plot_patients_generic(
        self,
        case,
        data,
        patients_idx="all",
        individual_parameters=None,
        reparametrized_ages=False,
        **kwargs,
    ):
        # plot with reparametrized ages
        ip_df, ip_torch = None, None
        if individual_parameters is not None:
            self._raise_if_model_not_init()
            ip_df, ip_torch = self._get_ip_df_torch(individual_parameters)

        # ---- Input manager
        plot_kws = self._plot_kwargs(case, kwargs)
        with_model = "model" in plot_kws  # plot reconstruction of model as well
        with_obs = "obs" in plot_kws and plot_kws["obs"].get("marker") is not None
        if not (with_model or with_obs):  # (or both !)
            raise LeaspyInputError(
                "Nothing to plot... nor model values nor observations."
            )

        # ---- Patients sublist
        if "patient_IDs" in kwargs.keys():
            warnings.warn(
                "Keyword argument <patient_IDs> is deprecated! "
                "Use <patients_idx> instead.",
                DeprecationWarning,
            )
            patients_idx = kwargs.get("patient_IDs")

        if isinstance(patients_idx, str):
            if patients_idx == "all":
                patients_idx = list(data.iter_to_idx.values())
            else:
                patients_idx = [patients_idx]

        # features check
        if self.model.is_initialized:
            if data.headers != self.model.features:
                raise LeaspyInputError(
                    "Features provided mismatch between data and model: "
                    f"{data.headers} != {self.model.features}"
                )

        ax, features, features_ix, labels, colors = self._handle_kwargs_begin(
            kwargs, data.headers
        )

        # Data to dataframe (only selected patients)
        df = data.to_dataframe()
        df["ID"] = df["ID"].astype(
            str
        )  # needed because of IndividualParameters converting ID int -> str
        df = df.set_index("ID").loc[patients_idx]

        if reparametrized_ages:
            if ip_df is None:
                raise LeaspyInputError(
                    "You want to plot reparametrized ages (`reparametrized_ages=True`) but you did not provide any individual parameters "
                    "to do so (please use `individual_parameters` argument)."
                )
            t0 = self.model.parameters["tau_mean"].item()
            df = df.join(ip_df)
            # reparametrized ages
            df["TIME_reparam"] = np.exp(df["xi"]) * (df["TIME"] - df["tau"]) + t0

        # ---- Plot

        # plot observations (with reparametrized times or not)
        if with_obs:
            self._plot_observations(
                ax, df, features, colors, reparametrized_ages, plot_kws["obs"]
            )

        # plot reconstruction as well (model values)
        if with_model:
            if ip_torch is None:
                raise LeaspyInputError(
                    "Individual reconstruction need valid individual parameters."
                )
            self._plot_model_trajectories(
                ax,
                df,
                self.model,
                ip_torch,
                features_ix,
                colors,
                reparametrized_ages,
                plot_kws["model"],
                **kwargs,
            )

        # ---- Title & labels
        if with_obs:
            title = "Observations"
            if with_model:
                title += " and individual trajectories"
        else:  # only with_model
            title = "Individual trajectories"
        ax.set_title(title)

        if reparametrized_ages:
            ax.set_xlabel("Reparametrized age")
        else:
            ax.set_xlabel("Age")

        self._handle_kwargs_end(ax, kwargs, colors, labels)

        return ax

    @staticmethod
    def _plot_observations(ax, df, features, colors, reparametrized_ages, plot_kws):
        """
        Internal routine: plot individual observations

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
        df : :class:`pandas.DataFrame`
            Data to plot
        features : list[:obj:`str`]
            Which features to plot (subset of model features / data features)
        colors : list
            List of colors (associated to features selected), in order
        reparametrized_ages : bool
            Should we plot trajectories in reparam age or not?
        plot_kws : dict
            Plot kwargs
        """

        if reparametrized_ages:
            time_col = "TIME_reparam"
        else:
            time_col = "TIME"

        df_with_time = df.set_index(df[time_col].rename("T"), append=True).sort_index()
        df_with_time = df_with_time[features].dropna(
            how="all"
        )  # selected features only

        for ind_id, ind_df in df_with_time.groupby("ID"):
            for (ft_name, s_ind_ft), ft_color in zip(ind_df.items(), colors):
                s_ind_ft = s_ind_ft.dropna()

                # TODO? use a cycle of markers to better distinguish individuals?
                ax.plot(
                    s_ind_ft.reset_index("T")["T"],
                    s_ind_ft,
                    c=ft_color,
                    # label=ft_lbl, # legend is done afterwards
                    **plot_kws,
                )

    @staticmethod
    def _plot_model_trajectories(
        ax,
        df,
        model,
        individual_parameters,
        features_ix,
        colors,
        reparametrized_ages,
        plot_kws,
        **kwargs,
    ):
        """
        Internal routine: plot individual trajectories estimated by model

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
        df : :class:`pandas.DataFrame`
            Data (TODO: could be the MultiIndex [ID,TIME] instead...)
        individual_parameters : tuple[list, dict]
            <!> in pytorch dict format: tuple(indices:list, dict{ip_name: vals})
        features_ix : list[int]
            Which features to plot (order of features from model)
        colors : list
            List of colors (associated to features selected), in order
        reparametrized_ages : bool
            Should we plot trajectories in reparam age or not?
        plot_kws : dict
            Plot kwargs
        **kwargs
            * "factor_past", "factor_future": float (default 0.5)
                past/future padding to plot (as fraction of total follow-up duration of subjects)
            * "n_tpts": int (default 100)
                nb of tpts in trajectory
        """

        ip_indices, ip_torch = individual_parameters

        for ind_id, ind_df in df.groupby("ID"):
            ind_ix = ip_indices.index(ind_id)
            ind_ip = {pn: pv[ind_ix] for pn, pv in ip_torch.items()}  # torch compatible

            timepoints = ind_df[
                "TIME"
            ]  # <!> always real patient ages here (to compute)
            min_t, max_t = min(timepoints), max(timepoints)
            total_t = max_t - min_t

            timepoints = np.linspace(
                min_t - kwargs.get("factor_past", 0.5) * total_t,
                max_t + kwargs.get("factor_future", 0.5) * total_t,
                kwargs.get("n_tpts", 100),
            )
            t = torch.tensor(timepoints, dtype=torch.float32).unsqueeze(0)

            trajectory = model.compute_individual_trajectory(t, ind_ip).squeeze(0)

            # times to plot if reparametrized ages are wanted
            if reparametrized_ages:
                timepoints = (
                    (
                        model.time_reparametrization(
                            t=t, alpha=ind_ip["xi"].exp(), tau=ind_ip["tau"]
                        )
                        + model.parameters["tau_mean"].item()
                    )
                    .squeeze(0)
                    .cpu()
                    .numpy()
                )

            for ft_ix, ft_color in zip(features_ix, colors):
                ax.plot(
                    timepoints,
                    trajectory[:, ft_ix],
                    c=ft_color,
                    # label=ft_lbl,
                    **plot_kws,
                )

    def patient_observations(
        self, data, patients_idx="all", individual_parameters=None, **kwargs
    ):
        """
        Plot patient observations

        Parameters
        ----------
        data : :class:`.Data`
        patients_idx : 'all' (default), :obj:`str` or list[:obj:`str`]
            Patients to display (by their ID).
        individual_parameters : :class:`.IndividualParameters` or :class:`pandas.DataFrame` (as may be output by ip.to_dataframe()) or dict (Pytorch ip format), optional
            If not None, observations are plotted with respect to reparametrized ages.
        """

        return self._plot_patients_generic(
            "obs",
            data,
            patients_idx=patients_idx,
            individual_parameters=individual_parameters,
            reparametrized_ages=individual_parameters is not None,
            **kwargs,
        )

    def patient_observations_reparametrized(
        self, data, individual_parameters, patients_idx="all", **kwargs
    ):
        """
        Plot patient observations (reparametrized ages)

        """

        return self._plot_patients_generic(
            "obs",
            data,
            patients_idx=patients_idx,
            individual_parameters=individual_parameters,
            reparametrized_ages=True,
            **kwargs,
        )

    def patient_trajectories(
        self,
        data,
        individual_parameters,
        patients_idx="all",
        reparametrized_ages=False,
        **kwargs,
    ):
        """
        Plot patient observations together with model individual reconstruction

        Parameters
        ----------
        data : :class:`.Data`
        individual_parameters : :class:`.IndividualParameters` or :class:`pandas.DataFrame` (as may be output by ip.to_dataframe()) or dict (Pytorch ip format)
        patients_idx : 'all' (default), :obj:`str` or list[:obj:`str`]
            Patients to display (by their ID).
        reparametrized_ages : :obj:`bool` (default False)
            Should we plot trajectories in reparam age or not? to study source impact essentially
        **kwargs
            cf. :meth:`._plot_model_trajectories`
            In particular, pass marker=None if you don't want observations besides model
        """

        return self._plot_patients_generic(
            "recons",
            data,
            patients_idx=patients_idx,
            individual_parameters=individual_parameters,
            reparametrized_ages=reparametrized_ages,
            **kwargs,
        )



from abc import abstractmethod
from typing import Iterable, Optional, Union

import torch

from leaspy.exceptions import LeaspyIndividualParamsInputError, LeaspyModelInputError
from leaspy.io.data.dataset import Dataset
from leaspy.utils.typing import DictParams, DictParamsTorch, KwargsType
from leaspy.utils.weighted_tensor import TensorOrWeightedTensor, WeightedTensor
from leaspy.variables.specs import (
    LVL_FT,
    DataVariable,
    LatentVariableInitType,
    ModelParameter,
    NamedVariables,
    SuffStatsRO,
    SuffStatsRW,
)
from leaspy.variables.state import State

from .obs_models import ObservationModel
from .stateful import StatefulModel

__all__ = ["McmcSaemCompatibleModel"]


class McmcSaemCompatibleModel(StatefulModel):
    """Defines probabilistic models compatible with an MCMC SAEM estimation.

    Parameters
    ----------
    name : :obj:`str`
        The name of the model.

    obs_models : :class:`~leaspy.models.obs_models` or :class:`~typing.Iterable` [:class:`~leaspy.models.obs_models`]
        The noise model for observations (keyword-only parameter).

    fit_metrics : :obj:`dict`
        Metrics that should be measured during the fit of the model
        and reported back to the user.

    **kwargs
        Hyperparameters for the model

    Attributes
    ----------
    is_initialized : :obj:`bool`
        Indicates if the model is initialized.
    name : :obj:`str`
        The model's name.
    features : :obj:`list` [:obj:`str`]
        Names of the model features.
    parameters : :obj:`dict`
        Contains the model's parameters
    obs_models : :obj:`tuple` [:class:`~leaspy.models.obs_models`, ...]
        The observation model(s) associated to the model.
    fit_metrics : :obj:`dict`
        Contains the metrics that are measured during the fit of the model and reported to the user.
    _state : :class:`~leaspy.variables.state.State`
        Private instance holding all values for model variables and their derived variables.
    """

    def __init__(
        self,
        name: str,
        *,
        # TODO? if we'd allow to pass a state there should be a all bunch of checks I guess? only "equality" of DAG is OK?
        # (WIP: cf. comment regarding inclusion of state here)
        # state: Optional[State] = None,
        # TODO? Factory of `ObservationModel` instead? (typically one would need the dimension to instantiate the `noise_std` variable of the right shape...)
        obs_models: Union[ObservationModel, Iterable[ObservationModel]],
        fit_metrics: Optional[dict[str, float]] = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        if isinstance(obs_models, ObservationModel):
            obs_models = (obs_models,)
        self.obs_models = tuple(obs_models)
        # load hyperparameters
        # <!> some may still be missing at this point (e.g. `dimension`, `source_dimension`, ...)
        # (thus we sh/could NOT instantiate the DAG right now!)
        self._load_hyperparameters(kwargs)
        # TODO: dirty hack for now, cf. AbstractFitAlgo
        self.fit_metrics = fit_metrics

    @property
    def observation_model_names(self) -> list[str]:
        """Get the names of the observation models.

        Returns
        -------
        :obj:`list` [:obj:`str`] :
            The names of the observation models.
        """
        return [model.to_string() for model in self.obs_models]

    def has_observation_model_with_name(self, name: str) -> bool:
        """
        Check if the model has an observation model with the given name.
        Parameters
        ----------
        name : :obj:`str`
            The name of the observation model to check.

        Returns
        -------
        :obj:`bool`:
            True if the model has an observation model with the given name, False otherwise.
        """
        return name in self.observation_model_names

    def to_dict(self, **kwargs) -> KwargsType:
        """Export model as a dictionary ready for export.

        Returns
        -------
        :class:`~leaspy.utils.typing.KwargsType`
            The model instance serialized as a dictionary.
        """
        d = super().to_dict()
        d.update(
            {
                "obs_models": {
                    obs_model.name: obs_model.to_string()
                    for obs_model in self.obs_models
                },
                "fit_metrics": self.fit_metrics,  # TODO improve
            }
        )
        return d

    @abstractmethod
    def _load_hyperparameters(self, hyperparameters: KwargsType) -> None:
        """Load model's hyperparameters.

        Parameters
        ----------
        hyperparameters : :obj:`dict` [ :obj:`str`, :obj:`Any` ]
            Contains the model's hyperparameters.

        """

    @classmethod
    def _raise_if_unknown_hyperparameters(
        cls, known_hps: Iterable[str], given_hps: KwargsType
    ) -> None:
        """Check if the given hyperparameters are known for the model.

        Parameters
        ----------
        known_hps : :obj:`Iterable` [:obj:`str`]
            The known hyperparameters for the model.
        given_hps : :class:`~leaspy.utils.typing.KwargsType`
            The hyperparameters provided to the model.

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If any unknown hyperparameter is provided to the model.
        """
        # TODO: replace with better logic from GenericModel in the future
        unexpected_hyperparameters = set(given_hps.keys()).difference(known_hps)
        if len(unexpected_hyperparameters) > 0:
            raise LeaspyModelInputError(
                f"Only {known_hps} are valid hyperparameters for {cls.__qualname__}. "
                f"Unknown hyperparameters provided: {unexpected_hyperparameters}."
            )

    @abstractmethod
    def _audit_individual_parameters(
        self, individual_parameters: DictParams
    ) -> KwargsType:
        """
        Perform various consistency and compatibility (with current model) checks
        on an individual parameters dict and outputs qualified information about it.

        TODO? move to IndividualParameters class?

        Parameters
        ----------
        individual_parameters : :class:`~leaspy.utils.typing.DictParams`
            Contains some individual parameters.
            If representing only one individual (in a multivariate model) it could be:
                * {'tau':0.1, 'xi':-0.3, 'sources':[0.1,...]}

            Or for multiple individuals:
                * {'tau':[0.1,0.2,...], 'xi':[-0.3,0.2,...], 'sources':[[0.1,...],[0,...],...]}

            In particular, a sources vector (if present) should always be a array_like, even if it is 1D

        Returns
        -------
        ips_info : :class:`~leaspy.utils.typing.KwargsType`
            * ``'nb_inds'`` : :obj:`int` >= 0
                Number of individuals present.
            * ``'tensorized_ips'`` : :obj:`dict` [ :obj:`str`, :class:`torch.Tensor` ]
                Tensorized version of individual parameters.
            * ``'tensorized_ips_gen'`` : generator
                Generator providing tensorized individual parameters for
                all individuals present (ordered as is).
        Raises
        ------
        :exc:`.NotImplementedError`
        """
        raise NotImplementedError

    def _get_tensorized_inputs(
        self,
        timepoints: torch.Tensor,
        individual_parameters: DictParamsTorch,
        *,
        skip_ips_checks: bool = False,
    ) -> tuple[torch.Tensor, DictParamsTorch]:
        """Convert the timepoints and individual parameters to tensors.

        Parameters
        ----------
        timepoints : :obj:`torch.Tensor`
            Contains the timepoints (age(s) of the subject).

        individual_parameters : :class:`~leaspy.utils.typing.DictParamsTorch`
            Contains the individual parameters.

        skip_ips_checks : :obj:`bool` (default: ``False``)
            Flag to skip consistency/compatibility checks and tensorization
            of ``individual_parameters`` when it was done earlier (speed-up).

        Returns
        -------
        :obj:`tuple` [:class:`torch.Tensor`, :class:`~leaspy.utils.typing.DictParamsTorch`]
            The timepoints and individual parameters converted to tensors.

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If computation is tried on more than 1 individual.
        """
        from .utilities import tensorize_2D

        if not skip_ips_checks:
            individual_parameters_info = self._audit_individual_parameters(
                individual_parameters
            )
            individual_parameters = individual_parameters_info["tensorized_ips"]
            if (n_individual_parameters := individual_parameters_info["nb_inds"]) != 1:
                raise LeaspyModelInputError(
                    "Only one individual computation may be performed at a time. "
                    f"{n_individual_parameters} was provided."
                )
        # Convert the timepoints (list of numbers, or single number) to a 2D torch tensor
        timepoints = tensorize_2D(timepoints, unsqueeze_dim=0)  # 1 individual
        return timepoints, individual_parameters

    def _check_individual_parameters_provided(
        self, individual_parameters_keys: Iterable[str]
    ) -> None:
        """Check consistency of individual parameters keys provided.

        Parameters
        ----------
        individual_parameters_keys : :obj:`Iterable` [:obj:`str`]
            The keys of the individual parameters provided.

        Raises
        ------
        :exc:`.LeaspyIndividualParamsInputError`
            If any of the individual parameters keys are unknown or missing.
        """
        ind_vars = set(self.individual_variables_names)
        unknown_ips = set(individual_parameters_keys).difference(ind_vars)
        missing_ips = ind_vars.difference(individual_parameters_keys)
        errs = []
        if len(unknown_ips):
            errs.append(f"Unknown individual latent variables: {unknown_ips}")
        if len(missing_ips):
            errs.append(f"Missing individual latent variables: {missing_ips}")
        if len(errs):
            raise LeaspyIndividualParamsInputError(". ".join(errs))

    def compute_individual_trajectory(
        self,
        timepoints: list[float],
        individual_parameters: DictParams,
        *,
        skip_ips_checks: bool = False,
    ) -> torch.Tensor:
        """Compute scores values at the given time-point(s) given a subject's individual parameters.

        .. note::
            The model uses its current internal state.

        Parameters
        ----------
        timepoints : :obj:`scalar` or :obj:`array_like` [:obj:`scalar`] (:obj:`list`, :obj:`tuple`, :class:`numpy.ndarray`)
            Contains the age(s) of the subject.

        individual_parameters : :class:`~leaspy.utils.typing.DictParams`
            Contains the individual parameters.
            Each individual parameter should be a scalar or array_like.

        skip_ips_checks : :obj:`bool` (default: ``False``)
            Flag to skip consistency/compatibility checks and tensorization
            of ``individual_parameters`` when it was done earlier (speed-up).

        Returns
        -------
        :class:`torch.Tensor`
            Contains the subject's scores computed at the given age(s)
            Shape of tensor is ``(1, n_tpts, n_features)``.
        """
        self._check_individual_parameters_provided(individual_parameters.keys())
        timepoints, individual_parameters = self._get_tensorized_inputs(
            timepoints, individual_parameters, skip_ips_checks=skip_ips_checks
        )
        # TODO? ability to revert back after **several** assignments?
        # instead of cloning the state for this op?
        local_state = self.state.clone(disable_auto_fork=True)
        self._put_data_timepoints(local_state, timepoints)
        for (
            individual_parameter_name,
            individual_parameter_value,
        ) in individual_parameters.items():
            local_state[individual_parameter_name] = individual_parameter_value

        return local_state["model"]

    def compute_prior_trajectory(
        self,
        timepoints: torch.Tensor,
        prior_type: LatentVariableInitType,
        *,
        n_individuals: Optional[int] = None,
    ) -> TensorOrWeightedTensor[float]:
        """
        Compute trajectory of the model for prior mode or mean of individual parameters.

        Parameters
        ----------
        timepoints : :obj:`torch.Tensor` [1, n_timepoints]
            Contains the timepoints (age(s) of the subject).

        prior_type : :class:`~leaspy.variables.specs.LatentVariableInitType`
            The type of prior to use for the individual parameters.

        n_individuals : :obj:`int`, optional
            The number of individuals.

        Returns
        -------
        :class:`torch.Tensor` [1, n_timepoints, dimension]
            The group-average values at given timepoints.

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If `n_individuals` is provided but not a positive integer, or if it is provided while `prior_type` is not `LatentVariableInitType.PRIOR_SAMPLES`.
        """
        exc_n_ind_iff_prior_samples = LeaspyModelInputError(
            "You should provide n_individuals (int >= 1) if, "
            "and only if, prior_type is `PRIOR_SAMPLES`"
        )
        if n_individuals is None:
            if prior_type is LatentVariableInitType.PRIOR_SAMPLES:
                raise exc_n_ind_iff_prior_samples
            n_individuals = 1
        elif prior_type is not LatentVariableInitType.PRIOR_SAMPLES or not (
            isinstance(n_individuals, int) and n_individuals >= 1
        ):
            raise exc_n_ind_iff_prior_samples
        local_state = self.state.clone(disable_auto_fork=True)
        self._put_data_timepoints(local_state, timepoints)
        local_state.put_individual_latent_variables(
            prior_type, n_individuals=n_individuals
        )

        return local_state["model"]

    def compute_mean_traj(
        self, timepoints: torch.Tensor
    ) -> TensorOrWeightedTensor[float]:
        """Trajectory for average of individual parameters (not really meaningful for non-linear models).

        Parameters
        ----------
        timepoints : :obj:`torch.Tensor` [1, n_timepoints]

        Returns
        -------
        :class:`torch.Tensor` [1, n_timepoints, dimension]
            The group-average values at given timepoints.
        """
        # TODO/WIP: keep this in BaseModel interface? or only provide `compute_prior_trajectory`, or `compute_mode|typical_traj` instead?
        return self.compute_prior_trajectory(
            timepoints, LatentVariableInitType.PRIOR_MEAN
        )

    def compute_mode_traj(
        self, timepoints: torch.Tensor
    ) -> TensorOrWeightedTensor[float]:
        """Most typical individual trajectory.

        Parameters
        ----------
        timepoints : :obj:`torch.Tensor` [1, n_timepoints]

        Returns
        -------
        :class:`torch.Tensor` [1, n_timepoints, dimension]
            The group-average values at given timepoints.

        """
        return self.compute_prior_trajectory(
            timepoints, LatentVariableInitType.PRIOR_MODE
        )

    def compute_jacobian_tensorized(
        self,
        state: State,
    ) -> DictParamsTorch:
        """
        Compute the jacobian of the model w.r.t. each individual parameter, given the input state.

        This function aims to be used in :class:`.ScipyMinimize` to speed up optimization.

        Parameters
        ----------
        state : :class:`~leaspy.variables.state.State`
            Instance holding values for all model variables (including latent individual variables)

        Returns
        -------
        :class:`~leaspy.utils.typing.DictParamsTorch`
            Tensors are of shape ``(n_individuals, n_timepoints, n_features, n_dims_param)``.

        Raises
        ------
        :exc:`NotImplementedError`
        """
        # return {
        #     ip: state[f"model_jacobian_{ip}"]
        #     for ip in self.get_individual_variable_names()
        # }
        raise NotImplementedError("This method is currently not implemented.")

    @classmethod
    def compute_sufficient_statistics(cls, state: State) -> SuffStatsRW:
        """
        Compute sufficient statistics from state.

        Parameters
        ----------
        state : :class:`~leaspy.variables.state.State`

        Returns
        -------
        :class:`~leaspy.variables.specs.SuffStatsRW`
            Contains the sufficient statistics computed from the state.
        """
        suff_stats = {}
        for mp_var in state.dag.sorted_variables_by_type[ModelParameter].values():
            mp_var: ModelParameter  # type-hint only
            suff_stats.update(mp_var.suff_stats(state))

        # we add some fake sufficient statistics that are in fact convergence metrics (summed over individuals)
        # TODO proper handling of metrics
        # We do not account for regularization of pop. vars since we do NOT have true Bayesian priors on them (for now)
        for k in ("nll_attach", "nll_regul_ind_sum"):
            suff_stats[k] = state[k]
        suff_stats["nll_tot"] = (
            suff_stats["nll_attach"] + suff_stats["nll_regul_ind_sum"]
        )  # "nll_regul_all_sum"

        return suff_stats

    @classmethod
    def update_parameters(
        cls,
        state: State,
        sufficient_statistics: SuffStatsRO,
        *,
        burn_in: bool,
    ) -> None:
        """
        Update model parameters of the provided state.

        Parameters
        ----------
        cls : :class:`~leaspy.models.mcmc_saem_compatible.McmcSaemCompatibleModel`
            Model class to which the parameters belong.

        state : :class:`~leaspy.variables.state.State`
            Instance holding values for all model variables (including latent individual variables)

        sufficient_statistics : :class:`~leaspy.variables.specs.SuffStatsRO`
            Contains the sufficient statistics computed from the state.

        burn_in : :obj:`bool`
            If True, the parameters are updated in a burn-in phase.
        """
        # <!> we should wait before updating state since some updating rules may depending on OLD state
        # (i.e. no sequential update of state but batched updates once all updated values were retrieved)
        # (+ it would be inefficient since we could recompute some derived values between updates!)
        params_updates = {}
        for mp_name, mp_var in state.dag.sorted_variables_by_type[
            ModelParameter
        ].items():
            mp_var: ModelParameter  # type-hint only
            params_updates[mp_name] = mp_var.compute_update(
                state=state, suff_stats=sufficient_statistics, burn_in=burn_in
            )
        # mass update at end
        for mp, mp_updated_val in params_updates.items():
            state[mp] = mp_updated_val

    def get_variables_specs(self) -> NamedVariables:
        """Get the specifications of the variables used in the model.

        Returns
        -------
        :class:`~leaspy.variables.specs.NamedVariables`
            Specifications of the variables used in the model, including timepoints and observation models.
        """
        specifications = NamedVariables({"t": DataVariable()})
        single_obs_model = len(self.obs_models) == 1
        for obs_model in self.obs_models:
            specifications.update(
                obs_model.get_variables_specs(named_attach_vars=not single_obs_model)
            )
        return specifications

    @abstractmethod
    def put_individual_parameters(self, state: State, dataset: Dataset):
        """Put the individual parameters inside the provided state (in-place).

        Raises
        ------
        :exc:`NotImplementedError`

        """
        raise NotImplementedError()

    def _put_data_timepoints(
        self, state: State, timepoints: TensorOrWeightedTensor[float]
    ) -> None:
        """Put the timepoints variables inside the provided state (in-place).

        Parameters
        ----------
        state : :class:`~leaspy.variables.state.State`
            Instance holding values for all model variables (including latent individual variables), as well as:
            - timepoints : :class:`torch.Tensor` of shape (n_individuals, n_timepoints)

        timepoints : :class:`~leaspy.utils.weighted_tensor.WeightedTensor` or :class:`torch.Tensor`
            Contains the timepoints (age(s) of the subject).

        Raises
        ------
        :exc:`TypeError`
            If the provided timepoints are not of type :class:`torch.Tensor` or :class:`~leaspy.utils.weighted_tensor.WeightedTensor`.
        """
        # TODO/WIP: we use a regular tensor with 0 for times so that 'model' is a regular tensor
        # (to avoid having to cope with `StatelessDistributionFamily` having some `WeightedTensor` as parameters)
        if isinstance(timepoints, WeightedTensor):
            state["t"] = timepoints
        elif isinstance(timepoints, torch.Tensor):
            state["t"] = WeightedTensor(timepoints)
        else:
            raise TypeError(
                f"Time points should be either torch Tensors or WeightedTensors. "
                f"Instead, a {type(timepoints)} was provided."
            )

    def put_data_variables(self, state: State, dataset: Dataset) -> None:
        """Put all the needed data variables inside the provided state (in-place).

        Parameters
        ----------

        state : :class:`~leaspy.variables.state.State`
            Instance holding values for all model variables (including latent individual variables), as well as:
            - timepoints : :class:`torch.Tensor` of shape (n_individuals, n_timepoints)
        dataset : :class:`~leaspy.io.data.dataset.Dataset`
            The dataset containing the data to be put in the state.
        """
        self._put_data_timepoints(
            state,
            WeightedTensor(
                dataset.timepoints, dataset.mask.to(torch.bool).any(dim=LVL_FT)
            ),
        )
        for obs_model in self.obs_models:
            state[obs_model.name] = obs_model.getter(dataset)

    def reset_data_variables(self, state: State) -> None:
        """Reset all data variables inside the provided state (in-place).

        Parameters
        ----------
        state : :class:`~leaspy.variables.state.State`
            Instance holding values for all model variables (including latent individual variables), as well as:
            - timepoints : :class:`torch.Tensor` of shape (n_individuals, n_timepoints)
        """
        state["t"] = None
        for obs_model in self.obs_models:
            state[obs_model.name] = None



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, Tuple

import torch

from leaspy.exceptions import LeaspyModelInputError
from leaspy.variables.state import State

__all__ = [
    "AbstractSampler",
    "AbstractIndividualSampler",
    "AbstractPopulationSampler",
]


class AbstractSampler(ABC):
    """
    Abstract sampler class.

    Parameters
    ----------
    name : :obj:`str`
        The name of the random variable to sample.
    shape : :obj:`tuple` of :obj:`int`
        The shape of the random variable to sample.
    acceptation_history_length : :obj:`int` > 0 (default 25)
        Deepness (= number of iterations) of the history kept for computing the mean acceptation rate.
        (It is the same for population or individual variables.)

    Attributes
    ----------
    name : :obj:`str`
        Name of variable
    shape : :obj:`tuple` of :obj:`int`
        Shape of variable
    acceptation_history_length : :obj:`int`
        Deepness (= number of iterations) of the history kept for computing the mean acceptation rate.
        (Same for population or individual variables by default.)
    acceptation_history : :class:`torch.Tensor`
        History of binary acceptations to compute mean acceptation rate for the sampler in MCMC-SAEM algorithm.
        It keeps the history of the last `acceptation_history_length` steps.

    Raises
    ------
    :exc:`.LeaspyModelInputError`
    """

    def __init__(
        self,
        name: str,
        shape: Tuple[int, ...],
        *,
        acceptation_history_length: int = 25,
    ):
        self.name = name
        self.shape = shape
        self.acceptation_history_length = acceptation_history_length
        self.acceptation_history = torch.zeros(
            (self.acceptation_history_length, *self.shape_acceptation)
        )

    @property
    def ndim(self) -> int:
        """Number of dimensions."""
        return len(self.shape)

    @property
    @abstractmethod
    def shape_acceptation(self) -> Tuple[int, ...]:
        """
        Return the shape of acceptation tensor for a single iteration.

        Returns
        -------
        :obj:`tuple` of :obj:`int` :
            The shape of the acceptation history.
        """

    @abstractmethod
    def sample(
        self,
        state: State,
        *,
        temperature_inv: float,
    ) -> None:
        """
        Apply a sampling step

        <!> It will modify in-place the internal state, caching all intermediate values needed to efficient.

        Parameters
        ----------
        state : :class:`.State`
            Object containing values for all model variables, including latent variables
        temperature_inv : :obj:`float` > 0
            Inverse of the temperature used in tempered MCMC-SAEM
        """

    def _group_metropolis_step(self, alpha: torch.Tensor) -> torch.Tensor:
        """
        Compute the boolean acceptance decisions.

        Parameters
        ----------
        alpha : :class:`torch.FloatTensor` > 0

        Returns
        -------
        accepted : :class:`torch.Tensor[bool]`, same shape as `alpha`
            Acceptance decision (False or True).
        """
        accepted = torch.rand(alpha.shape) < alpha
        return accepted

    def _metropolis_step(self, alpha: float) -> bool:
        """
        Compute the Metropolis acceptance decision.

        If better (alpha>=1): accept
        If worse (alpha<1): accept with probability alpha

        <!> This function is critical for the reproducibility between machines.
        Different architectures might lead to different rounding errors on alpha
        (e.g: 1. - 1e-6 vs 1. + 1e-6). If we were to draw only for alpha < 1 (and not when alpha >= 1),
        then it would cause the internal seed of pytorch to change or not depending on the case
        which would lead to very different results afterwards (all the random numbers would be affected).

        Parameters
        ----------
        alpha : :obj:`float` > 0

        Returns
        -------
        :obj:`bool`
            Acceptance decision (False or True).
        """
        # Sample a realization from uniform law
        # Choose to keep iff realization is < alpha (probability alpha)
        # <!> Always draw a number even if it seems "useless" (cf. docstring warning)
        return torch.rand(()) < alpha

    def _update_acceptation_rate(self, accepted: torch.Tensor):
        """
        Update history of acceptation rates with latest accepted rates

        Parameters
        ----------
        accepted : :class:`torch.FloatTensor` (0. or 1.)

        Raises
        ------
        :exc:`.LeaspyModelInputError`
        """
        # Concatenate the new acceptation result at end of new one (forgetting the oldest acceptation rate)
        old_acceptation_history = self.acceptation_history[1:]
        self.acceptation_history = torch.cat(
            [old_acceptation_history, accepted.unsqueeze(0)]
        )


class AbstractIndividualSampler(AbstractSampler):
    """
    Abstract class for samplers of individual random variables.

    Parameters
    ----------
    name : :obj:`str`
        The name of the random variable to sample.
    shape : :obj:`tuple` of :obj:`int`
        The shape of the random variable to sample.
    n_patients : :obj:`int`
        Number of patients.
    acceptation_history_length : :obj:`int` > 0 (default 25)
        Deepness (= number of iterations) of the history kept for computing the mean acceptation rate.
        (It is the same for population or individual variables.)

    Attributes
    ----------
    name : :obj:`str`
        Name of variable
    shape : :obj:`tuple` of :obj:`int`
        Shape of variable
    n_patients : :obj:`int`
        Number of patients.
    acceptation_history_length : :obj:`int`
        Deepness (= number of iterations) of the history kept for computing the mean acceptation rate.
        (It is the same for population or individual variables.)
    acceptation_history : :class:`torch.Tensor`
        History of binary acceptations to compute mean acceptation rate for the sampler in MCMC-SAEM algorithm.
        It keeps the history of the last `acceptation_history_length` steps.
    """

    def __init__(
        self,
        name: str,
        shape: Tuple[int, ...],
        *,
        n_patients: int,
        acceptation_history_length: int = 25,
    ):
        self.n_patients = n_patients
        super().__init__(
            name, shape, acceptation_history_length=acceptation_history_length
        )

        # Initialize the acceptation history
        #if self.ndim != 1:
        #    raise LeaspyModelInputError("Dimension of individual variable should be 1")


class AbstractPopulationSampler(AbstractSampler):
    """
    Abstract class for samplers of population random variables.

    Parameters
    ----------
    name : :obj:`str`
        The name of the random variable to sample.
    shape : :obj:`tuple` of :obj:`int`
        The shape of the random variable to sample.
    acceptation_history_length : :obj:`int` > 0 (default 25)
        Deepness (= number of iterations) of the history kept for computing the mean acceptation rate.
        (It is the same for population or individual variables.)
    mask : :class:`torch.Tensor`, optional
        A binary (0/1) tensor indicating which elements to sample.
        Elements with value 1 (True) are included in the sampling; elements with 0 (False) are excluded.

    Attributes
    ----------
    name : :obj:`str`
        Name of variable
    shape : :obj:`tuple` of :obj:`int`
        Shape of variable
    acceptation_history_length : :obj:`int`
        Deepness (= number of iterations) of the history kept for computing the mean acceptation rate.
        (It is the same for population or individual variables.)
    acceptation_history : :class:`torch.Tensor`
        History of binary acceptations to compute mean acceptation rate for the sampler in MCMC-SAEM algorithm.
        It keeps the history of the last `acceptation_history_length` steps.
    mask : :class:`torch.Tensor` of :obj:`bool`, optional
        A binary (0/1) tensor indicating which elements to sample.
        Elements with value 1 (True) are included in the sampling; elements with 0 (False) are excluded.
    """

    def __init__(
        self,
        name: str,
        shape: Tuple[int, ...],
        *,
        acceptation_history_length: int = 25,
        mask: Optional[torch.Tensor] = None,
    ):
        super().__init__(
            name, shape, acceptation_history_length=acceptation_history_length
        )
        if self.ndim not in {1, 2}:
            # convention: shape of pop variable is 1D or 2D
            raise LeaspyModelInputError(
                "Dimension of population variable should be 1 or 2"
            )

        # TODO / WIP
        self.mask = mask
        if self.mask is not None:
            raise NotImplementedError(
                "WIP: Masked samplers are not supported yet with new variable interface / new weighted tensors "
                "--> cf. comments in LatentVariable & StatelessDistributionFamily"
            )
            if not isinstance(self.mask, torch.Tensor) or self.mask.shape != self.shape:
                raise LeaspyModelInputError(
                    f"Mask for sampler should be of shape {self.shape} but is of shape {self.mask.shape}"
                )
            self.mask = self.mask.to(torch.bool)



from enum import Enum

import numpy as np

from leaspy.io.data import Dataset
from leaspy.io.outputs.individual_parameters import IndividualParameters
from leaspy.models import ConstantModel
from leaspy.utils.typing import FeatureType

from ..base import AlgorithmName
from ..settings import AlgorithmSettings
from .base import PersonalizeAlgorithm

__all__ = ["ConstantPredictionAlgorithm"]


class PredictionType(str, Enum):
    LAST = "last"
    LAST_KNOWN = "last-known"
    MAX = "max"
    MEAN = "mean"


class ConstantPredictionAlgorithm(
    PersonalizeAlgorithm[ConstantModel, IndividualParameters]
):
    r"""ConstantPredictionAlgorithm is an algorithm that provides constant predictions.

    It is used with the :class:`~leaspy.models.ConstantModel`.

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        The settings of constant prediction algorithm. It supports the following  `prediction_type` values (str)::
            * ``'last'``: last value even if NaN,
            * ``'last_known'``: last non NaN value,
            * ``'max'``: maximum (=worst) value ,
            * ``'mean'``: average of values

        depending on features, the `last_known` / `max` value may correspond to different visits.
        For a given feature, value will be NaN if and only if all values for this feature are NaN.

    Raises
    ------
    :exc:`.LeaspyAlgoInputError`
        If any invalid setting for the algorithm
    """

    name: AlgorithmName = AlgorithmName.PERSONALIZE_CONSTANT
    deterministic: bool = True

    def __init__(self, settings: AlgorithmSettings):
        super().__init__(settings)
        self.prediction_type: PredictionType = PredictionType(
            settings.parameters["prediction_type"]
        )

    def _compute_individual_parameters(
        self, model: ConstantModel, dataset: Dataset, **kwargs
    ) -> IndividualParameters:
        # always overwrite model features (no fit process)
        # TODO? we could fit the model before, only to recover model features,
        #  and then check at personalize that is the same (as in others personalize algos...)
        # Always overwrite model features (no fit for constant model...)
        model.initialize(dataset)
        individual_parameters = IndividualParameters()
        for individual in range(dataset.n_individuals):
            idx = dataset.indices[individual]
            times = dataset.get_times_patient(individual)
            values = dataset.get_values_patient(individual).numpy()
            ind_ip = self._get_individual_last_values(
                times, values, features=model.features
            )
            individual_parameters.add_individual_parameters(str(idx), ind_ip)
        return individual_parameters

    def _get_individual_last_values(
        self, times: np.ndarray, values: np.ndarray, *, features: list[FeatureType]
    ):
        """Get individual last values.

        Parameters
        ----------
        times : :class:`numpy.ndarray` [float]
            shape (n_visits,)

        values : :class:`numpy.ndarray` [float]
            shape (n_visits, n_features)

        features : list[FeatureType]
            Feature names

        Returns
        -------
        dict[ft_name: str, constant_value_to_be_padded]
        """
        # return a dict with parameters names being features names
        return dict(zip(features, self._get_feature_values(times, values)))

    def _get_feature_values(self, times: np.ndarray, values: np.ndarray):
        if self.prediction_type == PredictionType.MAX:
            return np.nanmax(values, axis=0)
        if self.prediction_type == PredictionType.MEAN:
            return np.nanmean(values, axis=0)
        sorted_indices = sorted(range(len(times)), key=times.__getitem__, reverse=True)
        # Sometimes, last value can be a NaN.
        # If this behavior is intended, then return it anyway
        if self.prediction_type == PredictionType.LAST:
            return values[sorted_indices[0]]
        values_sorted_desc = values[sorted_indices]
        # get first index of values being non nan, with visits ordered by more recent
        last_non_nan_ix_per_ft = (~np.isnan(values_sorted_desc)).argmax(axis=0)
        # 1 feature value will be nan iff feature was nan at all visits
        return values_sorted_desc[last_non_nan_ix_per_ft, range(values.shape[1])]



import warnings
from typing import Optional

import numpy as np
import pandas as pd

from leaspy.exceptions import LeaspyDataInputError
from leaspy.utils.typing import FeatureType

from .abstract_dataframe_data_reader import AbstractDataframeDataReader
from .individual_data import IndividualData
from .visit_dataframe_data_reader import VisitDataframeDataReader

__all__ = ["CovariateDataframeDataReader"]


class CovariateDataframeDataReader(AbstractDataframeDataReader):
    """
    Methods to convert :class:`pandas.DataFrame` to `Leaspy`-compliant data containers for longitudinal data with covariates.

    Parameters
    ----------
    covariate_names: List[str]
        Names of the columns in dataframe that contains the covariates

    Raises
    ------
    :exc:`.LeaspyDataInputError`
    """

    def __init__(
        self,
        *,
        covariate_names: list[str],
    ):
        super().__init__()
        if not covariate_names:
            raise LeaspyDataInputError("You must prrovide at least one covariate name.")
        self.covariate_names = covariate_names
        self.visit_reader = VisitDataframeDataReader()

    @property
    def long_outcome_names(self) -> list[FeatureType]:
        """Name of the longitudinal outcomes in dataset"""
        return self.visit_reader.long_outcome_names

    @property
    def n_visits(self) -> int:
        """Number of visit in the dataset"""
        return self.visit_reader.n_visits

    ######################################################
    #               COVARIATE METHODS
    ######################################################

    def _check_headers(self, columns: list[str]) -> None:
        """
        Check mendatory dataframe headers

        Parameters
        ----------
        columns: List[str]
            Names of the columns headers of the dataframe that contains patients information
        """
        self.visit_reader._check_headers(columns)

    def _set_index(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Set the index suited for the type of information contained in the dataframe

        Parameters
        ----------
        df: pd.DataFrame
            Dataframe with patient information

        Returns
        -------
        df: pd.DataFrame
            Dataframe with the right index
        """

        return self.visit_reader._set_index(df)

    def _clean_dataframe_covariates(
        self, df: pd.DataFrame, *, drop_full_nan: bool, warn_empty_column: bool
    ) -> pd.DataFrame:
        """
        Clean the dataframe that contains patient information

        Parameters
        ----------
        df: pd.DataFrame
            Dataframe with patient information

        drop_full_nan: bool
            If set to True, raw full of nan are dropped

        warn_empty_column: bool
            If set to True, a warning is raise for columns full of nan


        Returns
        -------
        df: pd.DataFrame
            Dataframe with clean information
        """

        df_covariate = df.copy(deep=True)

        if not (df_covariate.columns == self.covariate_names).all():
            raise LeaspyDataInputError(
                f"The covariate column names {df_covariate.columns} are "
                f"different from the provided covariate names {self.covariate_names}."
            )

        for covariate in self.covariate_names:
            if df_covariate[covariate].isna().any():
                raise LeaspyDataInputError(
                    f"Covariate '{covariate}' contains missing values (NaN)."
                    "Please ensure that values are provided for each visit."
                )

        for covariate in self.covariate_names:
            if not np.array_equal(
                df_covariate[covariate], df_covariate[covariate].astype(int)
            ):
                raise LeaspyDataInputError(
                    f"Covariate '{covariate}' must contain only integer values."
                )
            df_covariate[covariate] = df_covariate[covariate].astype(int)

        # Assert one unique covariate per patient and group to drop duplicates
        if (
            not (df_covariate.groupby("ID").nunique()[self.covariate_names].eq(1))
            .all()
            .all()
        ):
            raise LeaspyDataInputError(
                "There must be only an unique covariate value per patient."
            )
        df_covariate = df_covariate.groupby("ID").first()

        if len(df_covariate) == 0:
            raise LeaspyDataInputError("Dataframe should have at least 1 covariate")

        # Assert at least 2 different values per covariate
        for covariate in self.covariate_names:
            if (n_value := df_covariate[covariate].nunique(dropna=False)) < 2:
                raise LeaspyDataInputError(
                    f"The covariate '{covariate}' has only {n_value} unique value."
                    "Each covariate must have at least two distinct values across patients"
                )

        return df_covariate

    def _clean_dataframe(
        self, df: pd.DataFrame, *, drop_full_nan: bool, warn_empty_column: bool
    ) -> pd.DataFrame:
        """
        Clean the dataframe that contains patient information

        Parameters
        ----------
        df: pd.DataFrame
            Dataframe with patient information

        drop_full_nan: bool
            If set to True, raw full of nan are dropped

        warn_empty_column: bool
            If set to True, a warning is raise for columns full of nan


        Returns
        -------
        df: pd.DataFrame
            Dataframe with clean information
        """

        df_visit = self.visit_reader._clean_dataframe(
            df.drop(columns=self.covariate_names),
            drop_full_nan=drop_full_nan,
            warn_empty_column=warn_empty_column,
        )

        df_covariate = self._clean_dataframe_covariates(
            df.reset_index()
            .drop(self.long_outcome_names + ["TIME"], axis=1)
            .set_index("ID"),
            drop_full_nan=drop_full_nan,
            warn_empty_column=warn_empty_column,
        )

        if (
            not df_covariate.groupby("ID")
            .first()
            .index.equals(df_visit.groupby("ID").first().index)
        ):
            raise LeaspyDataInputError(
                "All patients must have at least one visit and one covariate"
            )

        df = df_visit.join(df_covariate)

        return df

    def _load_individuals_data_covariates(
        self, subj: IndividualData, df_subj: pd.DataFrame
    ) -> None:
        """
        Convert information stored in a dataframe to information stored into IndividualData

        Parameters
        ----------
        subj: IndividualData
            One patient with her/his information, potentially empty

        df_subj: pd.DataFrame
            One patient with her/his information
        """
        subj.add_covariates(
            covariates=df_subj[self.covariate_names].iloc[0].values.tolist()
        )

    def _load_individuals_data(
        self, subj: IndividualData, df_subj: pd.DataFrame
    ) -> None:
        """
        Convert information stored in a dataframe to information stored into IndividualData

        Parameters
        ----------
        subj: IndividualData
            One patient with her/his information, potentially empty

        df_subj: pd.DataFrame
            One patient with her/his information
        """
        self.visit_reader._load_individuals_data(subj, df_subj)
        self._load_individuals_data_covariates(subj, df_subj)



"""This module defines the State of stateful models.

A state contains 2 main components:
1. The relationships between the variables through a :class:`~leaspy.variables.dag.VariablesDAG`
2. The values of each variable of the DAG as a mapping between variable names and their values

The State class is crucial for stateful models with its logic for efficiently retrieving variable values. This class relies on a caching mechanism that enables quick queries.
"""

from __future__ import annotations

import copy
import csv
from collections.abc import MutableMapping
from contextlib import contextmanager
from enum import Enum, auto
from pathlib import Path
from typing import Iterable, Optional, Union

import pandas as pd
import torch

from leaspy.exceptions import LeaspyInputError
from leaspy.utils.weighted_tensor import WeightedTensor, unsqueeze_right

from .dag import VariablesDAG
from .specs import (
    Hyperparameter,
    IndividualLatentVariable,
    LatentVariableInitType,
    PopulationLatentVariable,
    VariableName,
    VariablesLazyValuesRO,
    VariablesLazyValuesRW,
    VariableValue,
)

__all__ = [
    "State",
    "StateForkType",
]


class StateForkType(Enum):
    """
    The strategy used to cache forked values in :class:`.State`.

    REF : Reference-based caching
        Cached values are stored by reference, meaning no copying occurs. Mutating the original
        variables after caching will affect the cached version.

    COPY : Deep copy-based caching
        Cached values are stored via `copy.deepcopy`, ensuring they are independent of the originals.

    Notes
    -----
    - Use `REF` for efficiency when you're certain the original values will not be mutated after caching.
    - Use `COPY` to ensure isolation between the cached values and any subsequent modifications.
    - If using `REF` beware that values will NOT be copied (it only keeps references of values),
    so do NOT mutate them directly or the behavior will be unexpected.
    """

    REF = auto()
    COPY = auto()

    def to_cache(self, d: VariablesLazyValuesRW) -> VariablesLazyValuesRO:
        """Get the values to cache, depending on forking type."""
        if self is self.REF:
            return d
        return {k: copy.deepcopy(v) for k, v in d.items()}


class State(MutableMapping):
    """
    Dictionary of cached values corresponding to the stateless DAG instance.

   Parameters
    ----------
    dag : :class:`~leaspy.variables.dag.VariablesDAG`
        The stateless DAG which state will hold values for.

    auto_fork_type : :class:`~leaspy.variables.state.StateForkType` or None (default)
        Refer to :class:`~leaspy.variables.state.StateForkType` class and :attr:`auto_fork_type`

    Attributes
    ----------
    dag : :class:`~leaspy.variables.dag.VariablesDAG`
        The stateless DAG which the state instance will hold values for.

    auto_fork_type : :class:`~leaspy.variables.state.StateForkType` or None
        If not `StateForkType.NONE` each dictionary assignment will lead to the partial caching
        of previous value and all its children, so they can be reverted without computation.
        The exact caching strategy depends on flag (caching by reference or by copy)
        Can be manually set or via `auto_fork` context manager.

    _tracked_variables : :ob:`set[:class:`~leaspy.variables.specs.VariableName`, ...]

    _values : :class:`~leaspy.variables.specs.VariablesLazyValuesRW`
        Private cache for values (computations are lazy thus some values may be None).
        All not None values are always self-consistent with respect to DAG dependencies.

    _last_fork : None or Optional[:class:`~leaspy.variables.specs.VariablesLazyValuesRO`]
        If not None, holds the previous partial state values so they may be `.revert()`.
        Automatically populated on assignment operations as soon as `auto_fork_type` is not `NONE`.
        Example: if you set a new value for `a`, then value of `a` and of all its children just before assignment
        are held until either reversion or a new assignment.
    """

    def __init__(
        self, dag: VariablesDAG, *, auto_fork_type: Optional[StateForkType] = None
    ):
        self.dag = dag
        self.auto_fork_type = auto_fork_type
        self._tracked_variables: set[VariableName, ...] = set()
        self._values: VariablesLazyValuesRW = {}
        self._last_fork: Optional[VariablesLazyValuesRO] = None
        self.clear()

    @property
    def tracked_variables(self) -> set[VariableName, ...]:
        """
        Get the set of variable names currently tracked by the State.

        Returns
        -------
        ;obj:`set`[:class:`~leaspy.variables.specs.VariableName`, ...]
            A set containing the names of the tracked variables.
        """
        return self._tracked_variables

    def track_variables(self, variable_names: Iterable[VariableName]) -> None:
        """
        Add some variables to the tracked variables.

        Parameters
        ----------
        variable_names : :obj:`~typing.Iterable` of :class:`~leaspy.variables.specs.VariableName`
            The names of the variables to be added to the tracked variables.
        """
        for variable_name in variable_names:
            self.track_variable(variable_name)

    def track_variable(self, variable_name: VariableName) -> None:
        """
        Add a single variable to the tracked variables.

        Parameters
        ----------
        variable_name : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable to be added to the tracked variables.
        """
        if variable_name in self.dag:
            self._tracked_variables.add(variable_name)

    def untrack_variables(self, variable_names: Iterable[VariableName]) -> None:
        """
        Remove some variables from the tracked variables.

        Parameters
        ----------
        variable_names : :obj:`~typing.Iterable` of :class:`~leaspy.variables.specs.VariableName`
            The names of the variables to be removed from the tracked variables.
        """
        for variable_name in variable_names:
            self.untrack_variable(variable_name)

    def untrack_variable(self, variable_name: VariableName) -> None:
        """
        Remove a single variable from the tracked variables.

        Parameters
        ----------
        variable_name : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable to be removed from the tracked variables.
        """
        if variable_name in self.dag:
            self._tracked_variables.discard(variable_name)

    def clear(self) -> None:
        """Reset last forked state and reset all values to their canonical values."""
        self._values = {
            n: var.value if isinstance(var, Hyperparameter) else None
            for n, var in self.dag.items()
        }
        self._last_fork = None

    def clone(
        self, *, disable_auto_fork: bool = False, keep_last_fork: bool = False
    ) -> State:
        """
        Clone current state without copying the DAG.

        Parameters
        ----------
        disable_auto_fork : :obj:`bool`, optional
            Whether to allow auto-fork or not.
            Default=False.

        keep_last_fork : :obj:`bool`, optional
            Whether to keep the last fork or not.
            Default=False.

        Returns
        -------
        :class:`~leaspy.variables.state.State` :
            The new cloned state instance.
        """
        cloned = State(
            self.dag, auto_fork_type=None if disable_auto_fork else self.auto_fork_type
        )
        cloned._values = copy.deepcopy(self._values)
        cloned._tracked_variables = self._tracked_variables
        if keep_last_fork:
            cloned._last_fork = copy.deepcopy(self._last_fork)
        return cloned

    @contextmanager
    def auto_fork(self, type: Optional[StateForkType] = StateForkType.REF):
        """
        Provide a context manager interface with temporary `auto_fork_type` set to `type`.
        
        Parameters
        ----------
        type :  :class:`~leaspy.variables.state.StateForkType` or None, optional
            The temporary auto-forking strategy to use within the context.
            Defaults to `StateForkType.REF`.

        Yields
        ------
        None
            Control returns to the caller with the temporary forking strategy applied.
        """
        orig_auto_fork_type = self.auto_fork_type
        try:
            self.auto_fork_type = type
            yield
        finally:
            self.auto_fork_type = orig_auto_fork_type

    def __iter__(self):
        """
        Iterates on keys (.keys(), .values() and .items() methods are automatically provided by `MutableMapping`).
        
        Returns
        -------
        iterator
            An iterator over the variable names (keys) stored in the state.
        """
        return iter(self._values)

    def __len__(self) -> int:
        """
        Get number of variables.
        
        Returns
        -------
        :obj:`int`
            The number of variables in the state.
        """
        return len(self._values)

    def _check_key_exists(self, name: VariableName) -> None:
        """
        Verify that a variable name exists in the DAG.

        Parameters
        ----------
        name : :class:`~leaspy.variables.specs.VariableName`
            The variable name to check.

        Raises
        ------
        :exc:`LeaspyInputError`
            If the variable name is not found in the DAG.
        """
        if name not in self.dag:
            raise LeaspyInputError(f"'{name}' is not a valid variable")

    def _get_or_compute_and_cache(
        self,
        name: VariableName,
        *,
        force_computation: bool = False,
        why: str = " to proceed",
    ) -> VariableValue:
        """
        Retrieve the cached value of a variable (unless `force_computation` is True) or compute it, 
        assuming node exists and all its ancestors have cached values.
        
        Parameters
        ----------
        name : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable to retrieve or compute.
        force_computation : :obj:`bool`, optional
            If True, forces recomputation of the variable value even if cached.
            Default is False.
        why : :obj:`str,` optional
            Explanation string used in the error message if computation fails. 
            Default is `" to proceed"`.

        Returns
        -------
        value : :class:`~leaspy.variables.specs.VariableValue`
            The cached or newly computed value of the variable.

        Raises
        ------
       :exc:`LeaspyInputError`
            If the variable is independent and its value cannot be computed (i.e., is None),
            indicating it is required.
        """
        if not force_computation:
            if (value := self._values[name]) is not None:
                return value
        value = self.dag[name].compute(self._values)
        if value is None:
            raise LeaspyInputError(
                f"'{name}' is an independent variable which is required{why}"
            )
        self._values[name] = value
        return value

    def __getitem__(self, name: VariableName) -> VariableValue:
        """
        Retrieve the cached value of a variable or compute it and then cache it 
        (as well as all intermediate computations that were needed).
        
        Parameters
        ----------
        name : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable to retrieve.

        Returns
        -------
        :class:`~leaspy.variables.specs.VariableValue`
            The cached or newly computed value of the variable.

        Raises
        ------
        :exc:`LeaspyInputError`
            If the variable or any of its dependencies cannot be computed.
        """
        if (value := self._get_value_from_cache(name)) is not None:
            return value
        for parent in self.dag.sorted_ancestors[name]:
            self._get_or_compute_and_cache(parent, why=f" to get '{name}'")
        return self._get_or_compute_and_cache(name, force_computation=True)

    def __contains__(self, name: VariableName) -> bool:
        """
        Check whether a variable exists in the DAG.

        Parameters
        ----------
        name : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable to check.

        Returns
        -------
        :obj:`bool`
            True if the variable is part of the DAG, False otherwise.
        """
        return name in self.dag

    def _get_value_from_cache(self, name: VariableName) -> Optional[VariableValue]:
        """
        Get the value for the provided variable name from the cache. 
        Raise an error if the value is not in the DAG.

        Parameters
        ----------
        name : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable whose cached value is to be retrieved.

        Returns
        -------
        Optional[:class:`~leaspy.variables.specs.VariableValue`]
            The cached value of the variable, or `None` if it hasn't been computed.
        """
        self._check_key_exists(name)
        return self._values[name]

    def is_variable_set(self, name: VariableName) -> bool:
        """
        Returns True if the variable is in the DAG and if its value is not None.

        Parameters
        ----------
        name : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable to check.

        Returns
        -------
        :obj:`bool`
            True if the variable exists in the DAG and its value has been set (i.e., is not None).
            False otherwise.
        """
        return self._get_value_from_cache(name) is not None

    def are_variables_set(self, variable_names: Iterable[VariableName]) -> bool:
        """
        Returns True if all the variables are in the DAG with values different from None.
        
        Parameters
        ----------
        variable_names : :obj:`Iterable`[:class:`~leaspy.variables.specs.VariableName`]
            A collection of variable names to check.

        Returns
        -------
        :obj:`bool`
            True if all variables exist in the DAG and their values are set (i.e., not None).
            False otherwise.
        """
        return all(self.is_variable_set(name) for name in variable_names)

    def __setitem__(self, name: VariableName, value: Optional[VariableValue]) -> None:
        """
        Smart and protected assignment of a variable value.

        Parameters
        ----------
        name : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable to be set.
        value : :class:`~leaspy.variables.specs.VariableValue`
            The value of the variable to set.

        Raises
        ------
        :exc:`LeaspyInputError`
            If the variable does not exist in the DAG or is not settable.
        """
        self._check_key_exists(name)
        if not self.dag[name].is_settable:
            raise LeaspyInputError(f"'{name}' is not intended to be set")
        sorted_children = self.dag.sorted_children[name]
        # automatically fork partial state to easily revert it
        if self.auto_fork_type is not None:
            self._last_fork = self.auto_fork_type.to_cache(
                {child: self._values[child] for child in (name,) + sorted_children}
            )
        # TODO? we do not "validate" / "check" input data for now
        #  (it could be a stateless variable method) to remain light
        self._values[name] = value
        # we reset values of all children of the node we just assigned a value to
        # (we postpone the evaluation of their new values when they will really be needed)
        for child in sorted_children:
            self._values[child] = None

    def put(
        self,
        variable_name: VariableName,
        variable_value: VariableValue,
        *,
        indices: tuple[int, ...] = (),
        accumulate: bool = False,
    ) -> None:
        """
        Smart and protected assignment of a variable value, but potentially on a subset of indices, 
        adding (accumulating) values and OUT-OF-PLACE.

        Parameters
        ----------
        variable_name : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable.
        variable_value : :class:`~leaspy.variables.specs.VariableValue`
            The new value to put in the variable name.
        indices : :obj:`tuple` of :obj:`int`, optional
            If set, the operation will happen on a subset of indices.
            Default=()
        accumulate : :obj:`bool`, optional
            If set to True, the new variable value will be added
            to the old value. Otherwise, it will be assigned.
            Default=False
        """
        if indices == ():
            # `torch.index_put` is not working in this case.
            if not accumulate:
                self[variable_name] = variable_value
            else:
                self[variable_name] = self[variable_name] + variable_value
            return
        # For now: no optimization for partial indices operations
        torch_indices = tuple(map(torch.tensor, indices))
        self[variable_name] = self[variable_name].index_put(
            indices=torch_indices,
            values=variable_value,
            accumulate=accumulate,
        )

    def __delitem__(self, name: VariableName) -> None:
        """
        Prevent deletion of variables from the state.

        Parameters
        ----------
        name : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable to delete (attempted).

        Raises
        ------
        :exc:`NotImplementedError`
            Always raised to indicate that variable deletion is not allowed.
        """
        raise NotImplementedError("Key removal is not allowed")

    def precompute_all(self) -> None:
        """Pre-compute all values of the graph (assuming leaves already have valid values)."""
        for n in self.dag:
            self._get_or_compute_and_cache(n)

    # def reset_to_admissible(self) -> None:
    #    """Reset all standard variables to their frozen or admissible values and pre-compute all other variables (forked state is cleared)."""
    #    # TODO: more generic?
    #    self.clear()
    #    for n, var in self.dag.items():
    #        if isinstance(var, StandardVariable):
    #            self._values[n] = var.admissible_value
    #    self.precompute_all()

    def revert(
        self, subset: Optional[VariableValue] = None, *, right_broadcasting: bool = True
    ) -> None:
        """
        Revert state to previous forked state. Forked state is then reset.

        Parameters
        ----------
        subset : :class:`~leaspy.variables.specs.VariableValue` or None
            If not None, the reversion is only partial:
            * subset = True <=> revert previous state for those indices
            * subset = False <=> keep current state for those indices
            <!> User is responsible for having tensor values that are consistent with
            `subset` shape (i.e. valid broadcasting) for the forked node and all of its children.
           <!> When the current OR forked state is not set (value = None) on a particular node of forked DAG,
           then the reverted result is always None.
        right_broadcasting : :obj:`bool`, optional
            If True and if `subset` is not None, then the subset of indices to revert uses right-broadcasting,
            instead of the standard left-broadcasting.
            Default=True.

        Raises
        ------
        :exc:`LeaspyInputError`
            If no forked state exists to revert from (i.e., `.auto_fork()` context was not used).
        """
        if self._last_fork is None:
            raise LeaspyInputError(
                "No forked state to revert from, please use within `.auto_fork()` context, "
                "or set `.auto_fork_type` to  `StateForkType.REF` or `StateForkType.COPY`."
            )
        if subset is None:
            self._values.update(self._last_fork)
            self._last_fork = None
            return
        to_revert = subset.to(torch.bool)
        to_keep = ~to_revert
        for k, old_v in self._last_fork.items():
            cur_v = self._values[k]
            if old_v is None or cur_v is None:
                self._values[k] = None
            else:
                assert (
                    old_v.shape == cur_v.shape
                ), f"Bad shapes for {k}: {old_v.shape} != {cur_v.shape}"
                if right_broadcasting:
                    add_ndim = max(old_v.ndim - to_revert.ndim, 0)
                    self._values[k] = old_v * unsqueeze_right(
                        to_revert, ndim=add_ndim
                    ) + cur_v * unsqueeze_right(to_keep, ndim=add_ndim)
                else:
                    self._values[k] = old_v * to_revert + cur_v * to_keep
        self._last_fork = None

    def to_device(self, device: torch.device) -> None:
        """
        Move values to the specified device (in-place).

        Parameters
        ----------
        device : :class:`torch.device`
        """
        for k, v in self._values.items():
            if v is not None:
                self._values[k] = v.to(device=device)
        if self._last_fork is not None:
            for k, v in self._last_fork.items():
                if v is not None:
                    self._last_fork[k] = v.to(device=device)

    def put_population_latent_variables(
        self, method: Optional[Union[str, LatentVariableInitType]]
    ) -> None:
        """"
        Initialize all population latent variables in the state with predefined values.

        Parameters
        ----------
        method : obj:`str` or :class:`~leaspy.variables.specs.LatentVariableInitType` or None
            The method used to initialize the variables. If None, all population latent variables
            will be unset (set to None). Otherwise, the corresponding initialization function will
            be called for each variable using the provided method.

        """
        # Nota: fixing order of variables in this loop is pointless since no randomness is involved in init of pop. vars
        for pp, var in self.dag.sorted_variables_by_type[
            PopulationLatentVariable
        ].items():
            var: PopulationLatentVariable  # for type-hint only
            if method is None:
                self[pp] = None
            else:
                self[pp] = var.get_init_func(method).call(self)

    def put_individual_latent_variables(
        self,
        method: Optional[Union[str, LatentVariableInitType]] = None,
        *,
        n_individuals: Optional[int] = None,
        df: Optional[pd.DataFrame] = None,
    ) -> None:
        """
        Initialize all individual latent variables in the state with predefined values.
        
        Parameters
        ----------
        method : :obj:`str` or  :class:`~leaspy.variables.specs.LatentVariableInitType`, optional
            The method used to initialize the variables. If None, the variables will be unset (set to None).
            If provided, an initialization function will be called per variable.
            When `method` is not None, `n_individuals` must be specified.
        n_individuals : :obl:`int`, optional
            Number of individuals to initialize. Required when `method` is not None and `df` is not provided.
        df : :obj:`pandas.DataFrame`, optional
            A DataFrame from which to directly extract the individual latent variable values.
            It must contain columns named 'tau' and 'xi' for direct assignment of these variables.
            If the "sources" variable is present, the DataFrame should include columns named
            'sources_0', 'sources_1', ..., up to the expected number of source variables.

        Raises
        ------
        :exc:`LeaspyInputError`
            If `method` is specified without `n_individuals`, or if required columns are missing in `df`.
        """
        if method is not None and n_individuals is None:
            raise LeaspyInputError(
                "`n_individuals` should not be None when `method` is not None."
            )
        individual_parameters = sorted(
            list(set(self.dag.sorted_variables_by_type[IndividualLatentVariable]))
        )
        if df is not None:
            for individual_parameter in individual_parameters:
                if individual_parameter in ("tau", "xi"):
                    self[individual_parameter] = torch.tensor(
                        df[[individual_parameter]].values, dtype=torch.double
                    )
                else:
                    nb_sources = len(df.columns) - 2
                    list_sources_name = [f"sources_{i}" for i in range(nb_sources)]
                    if not set(list_sources_name) <= set(df.columns):
                        raise LeaspyInputError(
                            "Please provide only individual parameters columns with sources stored as "
                            f"{list_sources_name}"
                        )
                    self["sources"] = torch.tensor(
                        df[list_sources_name].values, dtype=torch.double
                    ).float()
        else:
            for individual_parameter in individual_parameters:
                var: IndividualLatentVariable = self.dag[
                    individual_parameter
                ]  # for type-hint only
                if method is None:
                    self[individual_parameter] = None
                else:
                    self[individual_parameter] = var.get_init_func(
                        method, n_individuals=n_individuals
                    ).call(self)

    def save(self, output_folder: str, iteration: Optional[int] = None) -> None:
        """
        Save the tracked variable values of the state.

        Parameters
        ----------
        output_folder : :obj:`str`
            The path to the output folder in which the state's tracked variables should be saved.
        iteration : :obj:`int`, optional
            The iteration number when this method is called from an algorithm.
            This iteration number will appear at the beginning of the row.
        """
        output_folder = Path(output_folder)
        for variable in self._tracked_variables:
            dict_value = self._get_value_as_dict_of_lists(variable)
            for variable_name, value in dict_value.items():
                if iteration is not None:
                    value.insert(0, iteration)
                with open(
                    output_folder / f"{variable_name}.csv", "a", newline=""
                ) as filename:
                    writer = csv.writer(filename)
                    writer.writerow(value)

    def _get_value_as_dict_of_lists(
        self, variable_name: VariableName
    ) -> dict[VariableName, list[float]]:
        """
        Return the value of the given variable as a dictionary with list of floats.
        
        Parameters
        ----------
        variable_name : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable whose value is to be converted.

        Returns
        -------
        :obj:`dict` of {:class:`~leaspy.variables.specs.VariableName`: :obj:`list` of :obj:`float`}
            A dictionary mapping variable names (or indexed variable names for 2D tensors) to
            lists of floats representing the variable's values.

        Raises
        ------
        :exc:`ValueError`
            If the variable value tensor has more than 2 dimensions, which is unsupported.
        """
        value = self.__getitem__(variable_name)
        if isinstance(value, WeightedTensor):
            value = value.weighted_value
        if value.ndim == 0:
            return {variable_name: [value.tolist()]}
        elif value.ndim == 1:
            return {variable_name: value.tolist()}
        elif value.ndim == 2:
            if value.shape[1] == 1:
                return {variable_name: value[:, 0].tolist()}
            else:
                return {
                    f"{variable_name}_{i}": value[:, i].tolist()
                    for i in range(value.shape[1])
                }
        else:
            raise ValueError(
                f"The value of variable {variable_name} "
                f"is a tensor of dimension {value.ndim} > 2 "
                f"which is not supported. The value is: {value}."
            )

    def get_tensor_value(self, variable_name: VariableName) -> VariableValue:
        """
        Return the value of the provided variable as a torch tensor.

        Parameters
        ----------
        variable_name : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable for which to retrieve the value.

        Returns
        -------
        :class:`~leaspy.variables.specs.VariableValue` :
            The value of the variable.
        """
        if isinstance(self[variable_name], WeightedTensor):
            return self[variable_name].weighted_value
        return self[variable_name]

    def get_tensor_values(
        self, variable_names: Iterable[VariableName]
    ) -> tuple[VariableValue, ...]:
        """
        Return the values of the provided variables as torch tensors.

        Parameters
        ----------
        variable_names : :class:`~typing.Iterable` of :class:`~leaspy.variables.specs.VariableName`
            The names of the variables for which to retrieve the values.

        Returns
        -------
        :obj:`tuple` of :class:`~leaspy.variables.specs.VariableValue` :
            The values of the variables.
        """
        return tuple(self.get_tensor_value(name) for name in variable_names)
